import copy
import io
import logging
import time
from dataclasses import dataclass
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Union

if TYPE_CHECKING:
    import graphviz

import kumoapi.data_snapshot as snapshot_api
import kumoapi.graph as api
from kumoapi.common import JobStatus
from kumoapi.data_snapshot import GraphSnapshotID
from tqdm.auto import tqdm
from typing_extensions import Self

from kumoai import global_state
from kumoai.client.graph import GraphID
from kumoai.graph.table import Table
from kumoai.mixin import CastMixin

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL_S = 20


@dataclass(frozen=True)
class Edge(CastMixin, api.Edge):
    r"""An edge represents a relationship between two tables in a
    :class:`~kumoai.graph.Graph`. Note that edges are **always** bidirectional
    within the Kumo platform.

    Args:
        src_table: The name of the source table of the edge. This table must
            have a foreign key with name :obj:`fkey` that links to the primary
            key in the destination table.
        fkey: The name of the foreign key in the source table.
        dst_table: The name of the destination table in the graph. This table
            must have a primary key that links to the
            source table's foreign key.

    Example:
        >>> import kumoai
        >>> edge = kumoai.Edge("table_with_fkey", "fkey", "table_with_pkey")
    """
    def __iter__(self) -> Iterator[str]:
        # Allows unwrapping an edge via `src_table, fkey, dst_table = edge`
        return iter((self.src_table, self.fkey, self.dst_table))

    def __hash__(self) -> int:
        return hash((self.src_table, self.fkey, self.dst_table))

    @property
    def _fully_qualified_name(self) -> str:
        return f"{self.src_table}.{self.fkey}.{self.dst_table}"


@dataclass
class GraphHealthStats:
    r"""Graph health statistics contain important statistics that represent the
    healthiness of each defined edge in a graph. These statistics are computed
    as part of a :class:`~kumoai.graph.Graph` snapshot, and can be fetched by
    indexing with an :class:`~kumoai.graph.graph.Edge` object.
    """
    _stats: Dict[str, api.EdgeHealthStatistics]

    def __init__(self, stats: Dict[str, api.EdgeHealthStatistics]):
        self._stats = stats

    def __getitem__(self, key: Edge) -> api.EdgeHealthStatistics:
        return self._stats[key._fully_qualified_name]

    def __repr__(self) -> str:
        representation = "GraphHealthStats\n"
        for key, stats in self._stats.items():
            src, fkey, dst = key.split('.')
            representation += (f" - Edge({src} ({fkey})-> {dst}) \n")
            representation += (f"    - {stats.total_num_edges} total edges\n")
            representation += (
                f"    - {int(stats.absolute_match_stats.src_in_dst)} "
                f"({round(stats.percent_match_stats.src_in_dst, 2)}%) rows "
                f"in {src} have valid edges to {dst}\n")
            representation += (
                f"    - {int(stats.absolute_match_stats.dst_in_src)} "
                f"({round(stats.percent_match_stats.dst_in_src, 2)}%) rows "
                f"in {dst} have valid edges to {src}\n")
        return representation


class Graph:
    r"""A graph defines the relationships between a set of Kumo tables, akin
    to relationships between tables in a relational database. Creating a graph
    is the final step of data definition in Kumo; after a graph is created, you
    are ready to write a :class:`~kumoai.pquery.PredictiveQuery` and train a
    predictive model.


    .. code-block:: python

        import kumoai

        # Define connector to source data:
        connector = kumoai.S3Connector('s3://...')

        # Create Kumo Tables. See Table documentation for more information:
        customer = kumoai.Table(...)
        article = kumoai.Table(...)
        transaction = kumoai.Table(...)

        # Create a graph:
        graph = kumo.Graph(
            # These are the tables that participate in the graph: the keys of this
            # dictionary are the names of the tables, and the values are the Table
            # objects that correspond to these names:
            tables={
                'customer': customer,
                'stock': stock,
                'transaction': transaction,
            },

            # These are the edges that define the primary key / foreign key
            # relationships between the tables defined above. Here, `src_table`
            # is the table that has the foreign key `fkey`, which maps to the
            # table `dst_table`'s primary key:`
            edges=[
                dict(src_table='transaction', fkey='StockCode', dst_table='stock'),
                dict(src_table='transaction', fkey='CustomerID', dst_table='customer'),
            ],
        )

        # Validate the graph configuration, for use in Kumo downstream models:
        graph.validate(verbose=True)

        # Visualize the graph:
        graph.visualize()

        # Fetch the statistics of the tables in this graph (this method will
        # take a graph snapshot, and as a result may have high latency):
        graph.get_table_stats(wait_for="minimal")

        # Fetch link health statistics (this method will
        # take a graph snapshot, and as a result may have high latency):
        graph.get_edge_stats(non_blocking=Falsej)

    Args:
        tables: The tables in the graph, represented as a dictionary mapping
            unique table names (within the context of this graph) to the
            :class:`~kumoai.graph.Table` definition for the table.
        edges: The edges (relationships) between the :obj:`tables` in the
            graph. Edges must specify the source table, foreign key, and
            destination table that they link.

    .. # noqa: E501
    """
    def __init__(
        self,
        tables: Optional[Dict[str, Table]] = None,
        edges: Optional[List[Edge]] = None,
    ) -> None:
        self._tables: Dict[str, Table] = {}
        self._edges: List[Edge] = []

        for name, table in (tables or {}).items():
            self.add_table(name, table)

        for edge in (edges or []):
            self.link(Edge._cast(edge))

        # Cached from backend:
        self._graph_snapshot_id: Optional[GraphSnapshotID] = None

    def print_definition(self) -> None:
        r"""Prints the full definition for this graph; the definition uses
        placeholder names in place of `kumoai.graph.Table` variables. Copy and
        paste this definition, modify the table variable names to re-create
        the original graph.

        Example:
            >>> import kumoai
            >>> graph = kumoai.Graph(...)  # doctest: +SKIP
            >>> graph.print_definition()  # doctest: +SKIP
            Graph(
              tables={
                'table-1' : <table-1>,
                'table-2' : <table-2>,
                ...
                'table-N' : <table-N>,
              },
              edges=[
                Edge(src_table='table-A', fkey='fkey-AD', dst_table='table-D'),
                Edge(src_table='table-B', fkey='fkey-BE', dst_table='table-E'),
                ...
                Edge(src_table='table-C', fkey='fkey-CF', dst_table='table-F'),
              ],
            )
        """
        definition = f"{self.__class__.__name__}(\n"
        definition += "  tables={"
        for table in self._tables.keys():
            definition += f"\n    '{table}' : <{table}>,"
        definition += "\n  },\n"
        definition += "  edges=["
        for edge in self._edges:
            src_table, fkey, dst_table = edge
            definition += (
                f"\n    {edge.__class__.__name__}(src_table='{src_table}', "
                f"fkey='{fkey}', dst_table='{dst_table}'),")
        definition += "\n  ],\n)"
        print(definition)

    # Properties ##############################################################

    @property
    def id(self) -> str:
        r"""Returns the unique ID for this graph, determined from its
        schema and the schemas of the tables and columns that it contains. Two
        graphs with any differences in their constituent tables or columns are
        guaranteed to have unique identifiers.
        """
        return self.save()

    # Save / load #############################################################

    def _to_api_graph_definition(self) -> api.GraphDefinition:
        col_groups_by_dst_table: Dict[str, List[api.ColumnKey]] = dict()
        for edge in self.edges:
            dst_pkey = self[edge.dst_table].primary_key
            if dst_pkey is None:
                raise ValueError(
                    f"The destination table {edge.dst_table} of edge "
                    f"{edge} does not have a primary key.")
            if edge.dst_table not in col_groups_by_dst_table:
                col_groups_by_dst_table[edge.dst_table] = [
                    api.ColumnKey(edge.dst_table, dst_pkey.name)
                ]
            col_groups_by_dst_table[edge.dst_table].append(
                api.ColumnKey(edge.src_table, edge.fkey))

        return api.GraphDefinition(
            tables={
                table_name: table._to_api_table_definition()
                for table_name, table in self.tables.items()
            },
            col_groups=[
                api.ColumnKeyGroup(columns=tuple(col_keys))
                for col_keys in col_groups_by_dst_table.values()
            ],
        )

    @staticmethod
    def _edges_from_api_graph_definition(
            graph_definition: api.GraphDefinition) -> List[Edge]:
        edges: List[Edge] = []
        for col_group in graph_definition.col_groups:
            pkey_col = None
            for col in col_group.columns:
                table_def = graph_definition.tables[col.table_name]
                if col.col_name == table_def.pkey:
                    pkey_col = col
                    break
            assert pkey_col is not None
            for col in col_group.columns:
                if col != pkey_col:
                    edges.append(
                        Edge(src_table=col.table_name, fkey=col.col_name,
                             dst_table=pkey_col.table_name))

        return edges

    @staticmethod
    def _from_api_graph_definition(
            graph_definition: api.GraphDefinition) -> 'Graph':
        tables = {
            k: Table._from_api_table_definition(v)
            for k, v in graph_definition.tables.items()
        }
        edges = Graph._edges_from_api_graph_definition(graph_definition)
        return Graph(tables, edges)

    def save(self, name: Optional[str] = None,
             skip_validation: bool = False, ) -> Union[GraphID, str]:
        r"""Associates this graph with a unique name, that can later be
        used to fetch the graph either in the Kumo UI or in the Kumo SDK
        with method :meth:`~kumoai.Graph.load`.

        Args:
            name: The name to associate with this table definition. If the
                name is already associated with another table, that table will
                be overridden.
            skip_validation: Whether to skip validation of the graph. If
                :obj:`True`, validation will be skipped, but saving an invalid
                graph may result in undefined behavior.
                If :obj:`False`, the graph will be validated before saving.

        Example:
            >>> import kumoai
            >>> graph = kumoai.Graph(...)  # doctest: +SKIP
            >>> graph.save()  # doctest: +SKIP
            graph-xxx
            >>> graph.save("template_name")  # doctest: +SKIP
            >>> loaded = kumoai.Graph.load("template_name")  # doctest: +SKIP
        """
        if not skip_validation:
            self.validate(verbose=False)

        template_resource = (global_state.client.graph_api.get_graph_if_exists(
            graph_id_or_name=name)) if name else None
        if template_resource is not None:
            config = self._from_api_graph_definition(template_resource.graph)
            logger.warning(
                ("Graph template %s already exists, with configuration %s. "
                 "This template will be overridden with configuration %s."),
                name, str(config), str(self))

        # Save as named template
        return global_state.client.graph_api.create_graph(
            graph_def=self._to_api_graph_definition(),
            force_rename=True if name else False,
            name_alias=name,
        )

    @classmethod
    def load(cls, graph_id_or_template: str) -> 'Graph':
        r"""Loads a graph from either a graph ID or a named template. Returns a
        :class:`Graph` object that contains the loaded graph along with its
        associated tables, columns, etc.
        """
        api = global_state.client.graph_api
        res = api.get_graph_if_exists(graph_id_or_template)
        if not res:
            raise ValueError(f"Graph {graph_id_or_template} was not found.")
        out = cls._from_api_graph_definition(res.graph)
        return out

    # Snapshot ################################################################

    @property
    def snapshot_id(self) -> Optional[snapshot_api.GraphSnapshotID]:
        r"""Returns the snapshot ID of this graph's snapshot, if a snapshot
        has been taken. Returns `None` otherwise.

        .. warning::
            This function currently only returns a snapshot ID if a snapshot
            has been taken *in this session.*
        """
        return self._graph_snapshot_id

    def snapshot(
        self,
        *,
        force_refresh: bool = False,
        non_blocking: bool = False,
    ) -> snapshot_api.GraphSnapshotID:
        r"""Takes a *snapshot* of this graph's underlying data, and returns a
        unique identifier for this snapshot.

        This is equivalent to taking a snapshot for each constituent table in
        the graph.  For more information, please see the documentation for
        :meth:`~kumoai.graph.Table.snapshot`.

        .. warning::
            Please familiarize yourself with the warnings for this method in
            :class:`~kumoai.graph.Table` before proceeding.

        Args:
            force_refresh: Indicates whether a snapshot should be taken, if one
                already exists in Kumo. If :obj:`False`, a previously existing
                snapshot may be re-used. If :obj:`True`, a new snapshot is
                always taken.
            non_blocking: Whether this operation should return immediately
                after creating the snapshot, or await completion of the
                snapshot. If :obj:`True`, the snapshot will proceed in the
                background, and will be used for any downstream job.

        Raises:
            RuntimeError: if ``non_blocking`` is set to :obj:`False` and the
                graph snapshot fails.
        """
        if self._graph_snapshot_id is None or force_refresh:
            self.save()
            if not force_refresh:
                snapshotted_table_names: List[str] = []
                for table_name, table in self.tables.items():
                    if table.snapshot_id is not None:
                        snapshotted_table_names.append(table_name)
                if len(snapshotted_table_names) > 0:
                    logger.warning(
                        "Tables %s have already been snapshot, and will not "
                        "be refreshed. If you would like to refresh all "
                        "tables, please set 'force_refresh=True'.",
                        snapshotted_table_names)

            self._graph_snapshot_id = (
                global_state.client.graph_api.create_snapshot(
                    graph_id=self.id,
                    refresh_source=True,
                ))
            logger.info("Graph snapshot with identifier %s created.",
                        self._graph_snapshot_id)

            # Perform initial GET to update table snapshot IDs:
            graph_resource: snapshot_api.GraphSnapshotResource = (
                global_state.client.graph_api.get_snapshot(
                    snapshot_id=self._graph_snapshot_id))
            for table_name, table_id in graph_resource.table_ids.items():
                self[table_name]._table_snapshot_id = table_id

            # NOTE we do not use a `KumoFuture` here as we do not want to treat
            # a graph refresh as having its own state; since we only ever
            # operate on the latest graph version (and do not let users to time
            # travel), there is no need for a separate Future object:
            if not non_blocking:
                stage = snapshot_api.GraphSnapshotStage.INGEST
                table_status: Dict[str, JobStatus] = {
                    table_name: JobStatus.NOT_STARTED
                    for table_name in self.tables
                }

                # Increment progress bar with table refresh stages:
                done = [status.is_terminal for status in table_status.values()]
                graph_done = False
                if logger.isEnabledFor(logging.INFO):
                    pbar = tqdm(total=len(done), unit="table",
                                desc="Ingesting")
                while not (all(done) and graph_done):
                    graph_resource = (
                        global_state.client.graph_api.get_snapshot(
                            snapshot_id=self._graph_snapshot_id))
                    for table_name, table_id in graph_resource.table_ids.items(
                    ):
                        resource = (global_state.client.table_api.get_snapshot(
                            snapshot_id=table_id))
                        table_status[table_name] = resource.stages[
                            stage].status
                    done = [
                        status.is_terminal for status in table_status.values()
                    ]
                    graph_done = graph_resource.stages[
                        stage].status.is_terminal
                    if logger.isEnabledFor(logging.INFO):
                        pbar.update(sum(done) - pbar.n)
                    time.sleep(_DEFAULT_INTERVAL_S)
                if logger.isEnabledFor(logging.INFO):
                    pbar.update(len(done) - pbar.n)
                    pbar.close()

                state = graph_resource.stages[stage]
                status = state.status
                warnings = "\n".join([
                    f"{i}. {message}"
                    for i, message in enumerate(state.warnings)
                ])
                error = state.error
                if status == JobStatus.FAILED:
                    raise RuntimeError(
                        f"Graph snapshot with identifier "
                        f"{self._graph_snapshot_id} failed, with error "
                        f"{error} and warnings {warnings}")
                if len(state.warnings) > 0:
                    logger.warning(
                        "Graph snapshot completed with the following "
                        "warnings: %s", warnings)
        else:
            logger.warning(
                "Graph snapshot with identifier %s already exists, and will "
                "not be refreshed.", self._graph_snapshot_id)

        # <prefix>@<data_version>:
        assert self._graph_snapshot_id is not None
        return self._graph_snapshot_id

    # Statistics ##############################################################

    def get_table_stats(
        self,
        wait_for: Optional[str] = None,
    ) -> Dict[str, Dict[str, Any]]:
        r"""Returns all currently computed statistics on the latest snapshot of
        this graph. If a snapshot on this graph has not been taken, this method
        will take a snapshot.

        .. note::
            Graph statistics are computed in multiple stages after ingestion is
            complete. These stages are called *minimal* and *full*; minimal
            statistics are always computed before full statistics.

        Args:
            wait_for: Whether this operation should block on the existence of
                statistics availability. This argument can take one of three
                values: :obj:`None`, which indicates that the method should
                return immediately with whatever statistics are present,
                :obj:`"minimal"`, which indicates that the method should return
                the when the minimum, maximum, and fraction of NA values
                statistics are present, or :obj:`"full"`, which indicates that
                the method should return when all computed statistics are
                present.
        """
        assert wait_for is None or wait_for in {"minimal", "full"}

        # Wait for graph ingestion to be done:
        if not self._graph_snapshot_id:
            self.snapshot(force_refresh=False, non_blocking=False)
        assert self._graph_snapshot_id is not None

        # Wait for all table snapshots to match the `wait_for` stage, if
        # we support that:
        if wait_for:
            if wait_for == "minimal":
                stage = snapshot_api.TableSnapshotStage.MIN_COL_STATS
            else:
                stage = snapshot_api.TableSnapshotStage.FULL_COL_STATS

            table_status: Dict[str, JobStatus] = {
                table_name: JobStatus.NOT_STARTED
                for table_name in self.tables
            }
            done = [status.is_terminal for status in table_status.values()]
            if logger.isEnabledFor(logging.INFO):
                pbar = tqdm(total=len(done), unit="table",
                            desc="Computing Statistics")
            while not all(done):
                for table_name, table in self.tables.items():
                    resource = (global_state.client.table_api.get_snapshot(
                        snapshot_id=table._table_snapshot_id))
                    table_status[table_name] = resource.stages[stage].status
                done = [status.is_terminal for status in table_status.values()]
                if logger.isEnabledFor(logging.INFO):
                    pbar.update(sum(done) - pbar.n)
                time.sleep(_DEFAULT_INTERVAL_S)
            if logger.isEnabledFor(logging.INFO):
                pbar.update(len(done) - pbar.n)
                pbar.close()

        # Write out statistics:
        out = {}
        for table_name, table in self.tables.items():
            resource = (global_state.client.table_api.get_snapshot(
                snapshot_id=table._table_snapshot_id))
            out[table_name] = {
                stat.column_name: stat.stats
                for stat in resource.column_stats
            }
        return out

    def get_edge_stats(
        self,
        *,
        non_blocking: bool = False,
    ) -> Optional[GraphHealthStats]:
        """Retrieves edge health statistics for the edges in a graph, if these
        statistics have been computed by a graph snapshot.

        Edge health statistics are returned in a
        :class:`~kumoai.graph.GraphHealthStats` object, and contain information
        about the match rate between primary key / foreign key relationships
        between the tables in the graph.

        Args:
            non_blocking: Whether this operation should return immediately
                after querying edge statistics (returning `None` if statistics
                are not available), or await completion of statistics
                computation.
        """
        if self._graph_snapshot_id is None:
            raise ValueError('In order to calculate edge health statistics, '
                             'you must first create a snapshot of the graph '
                             'on which to calculate match statistics for each '
                             'edge. Please call Graph.snapshot() and then '
                             'this function.')

        edge_health_response = global_state.client.graph_api.get_edge_stats(
            graph_snapshot_id=self._graph_snapshot_id)

        if non_blocking:
            if not edge_health_response.is_ready:
                return None
        else:
            while not edge_health_response.is_ready:
                edge_health_response = (
                    global_state.client.graph_api.get_edge_stats(
                        graph_snapshot_id=self._graph_snapshot_id))

        return GraphHealthStats(edge_health_response.statistics)

    # Tables ##################################################################

    def has_table(self, name: str) -> bool:
        r"""Returns True if a table by `name` is present in this Graph."""
        return name in self._tables

    def table(self, name: str) -> Table:
        r"""Returns a table in this Kumo Graph.

        Raises:
            KeyError: if no such table is present.
        """
        if name not in self._tables:
            raise KeyError(f"Table '{name}' not found in this graph.")
        return self._tables[name]

    def add_table(self, name: str, table: Table) -> 'Graph':
        r"""Adds a table to this Kumo Graph.

        Raises:
            KeyError: if a table with the same name already exists in this
                graph.
        """
        if name in self._tables:
            raise KeyError(
                f"Cannot add table with name '{name}' to this graph; names "
                f"must be globally unique within a graph.")
        self._tables[name] = table
        return self

    def remove_table(self, name: str) -> Self:
        r"""Removes a table from this graph.

        Raises:
            KeyError: if no such table is present.
        """
        if not self.has_table(name):
            raise KeyError(f"Table '{name}' not found in this graph.'")

        del self._tables[name]
        self._edges = [
            edge for edge in self._edges
            if edge.src_table != name and edge.dst_table != name
        ]
        return self

    @property
    def tables(self) -> Dict[str, Table]:
        r"""Returns a list of all :class:`~kumoai.graph.Table` objects that
        are contained in this graph.
        """
        return self._tables

    def infer_metadata(self, inplace: bool = True) -> 'Graph':
        r"""Infers metadata for the tables in this Graph, by inferring the
        metadata of each table in the graph. For more information, please
        see the documentation for
        :meth:`~kumoai.table.Table.infer_metadata`.
        """
        out = self
        if not inplace:
            out = copy.deepcopy(self)

        for table in out.tables.values():
            table.infer_metadata(inplace=True)
        return out

    # Edges ###################################################################

    def infer_links(self) -> 'Graph':
        r"""Infers edges for the tables in this Graph. It adds edges to the
        graph.

        Note that the function only works if the graph edges are empty.
        """
        if self._edges is not None and len(self._edges) > 0:
            raise ValueError(
                "Cannot infer links if graph edges are not empty.")

        graph_def_with_col_groups = global_state.client.graph_api.infer_links(
            graph=self._to_api_graph_definition())

        edges = Graph._edges_from_api_graph_definition(
            graph_def_with_col_groups)

        for edge in (edges or []):
            logger.info("Inferring edge: %s", edge)
            self.link(Edge._cast(edge))
        return self

    def link(self, *args: Optional[Union[str, Edge]],
             **kwargs: str) -> 'Graph':
        r"""Links two tables (:obj:`src_table` and :obj:`dst_table`) from the
        foreign key :obj:`fkey` in the source table to the primary key in the
        destination table. These edges are treated bidirectionally in Kumo.

        Args:
            *args: Any arguments to construct a
                :class:`kumoai.graph.Edge`, or a :class:`kumoai.graph.Edge`
                itself.
            **kwargs: Any keyword arguments to construct a
                :class:`kumoai.graph.Edge`.

        Raises:
            ValueError: if the edge is already present in the graph, if the
                source table does not exist in the graph, if the destination
                table does not exist in the graph, if the source key does not
                exist in the source table, or if the primary key of the source
                table is being treated as a foreign key.
        """
        edge = Edge._cast(*args, **kwargs)
        if edge is None:
            raise ValueError("Cannot add a 'None' edge to a graph.")

        (src_table, fkey, dst_table) = edge

        if edge in self._edges:
            raise ValueError(f"Cannot add edge {edge} to graph; edge is "
                             f"already present.")

        if src_table not in self._tables:
            raise ValueError(
                f"Source table '{src_table}' does not exist in the graph. "
                f"Please add it via `Graph.add_table(...)` before proceeding.")

        if dst_table not in self._tables:
            raise ValueError(
                f"Destination table '{dst_table}' does not exist in the "
                f"graph. Please add it via `Graph.add_table(...)` before "
                f"proceeding.")

        if fkey not in self._tables[src_table]:
            raise ValueError(
                f"Source key '{fkey}' does not exist in source table "
                f"'{src_table}'; please check that you have added it as a "
                f"column.")

        # Backend limitations: ensure the source is not its primary key:
        src_pkey = self.table(src_table).primary_key
        src_is_pkey = src_pkey is not None and src_pkey.name == fkey
        if src_is_pkey:
            raise ValueError(f"Cannot treat the primary key of table "
                             f"'{src_table}' as a foreign key; please "
                             f"select a different key.")

        self._edges.append(edge)
        return self

    def unlink(self, *args: Optional[Union[str, Edge]],
               **kwargs: str) -> 'Graph':
        r"""Removes an edge added to a Kumo Graph.

        Args:
            *args: Any arguments to construct a
                :class:`~kumoai.graph.Edge`, or a :class:`~kumoai.graph.Edge`
                itself.
            **kwargs: Any keyword arguments to construct a
                :class:`~kumoai.graph.Edge`.

        Raises:
            ValueError: if the edge is not present in the graph.
        """
        edge = Edge._cast(*args, **kwargs)
        if edge not in self._edges:
            raise ValueError(f"Edge {edge} is not present in {self._edges}")
        self._edges.remove(edge)
        return self

    @property
    def edges(self) -> List[Edge]:
        r"""Returns a list of all :class:`~kumoai.graph.Edge` objects that
        represent links in this graph.
        """
        return self._edges

    def validate(self, verbose: bool = True) -> Self:
        r"""Validates a Graph to ensure that all relevant metadata is specified
        for its Tables and Edges.

        Concretely, validation ensures that all tables are valid (see
        :meth:`~kumoai.graph.table.validate` for more information), and that
        edges properly link primary keys and foreign keys between valid tables.
        It additionally ensures that primary and foreign keys between tables
        in an edge are of the same data type, so that unexpected mismatches do
        not occur within the Kumo platform.

        Example:
            >>> import kumoai
            >>> graph = kumoai.Graph(...)  # doctest: +SKIP
            >>> graph.validate()  # doctest: +SKIP
            ValidationResponse(warnings=[], errors=[])

        Args:
            verbose: Whether to log non-error output of this validation.

        Raises:
            ValueError:
                if validation fails.
        """
        # Validate table definitions, so we can properly create a graph
        # definition:
        for table_name, table in self.tables.items():
            try:
                table.validate(verbose=verbose)
            except ValueError as e:
                raise ValueError(
                    f"Validation of table {table_name} failed. {e}") from e

        resp = global_state.client.graph_api.validate_graph(
            api.GraphValidationRequest(self._to_api_graph_definition()))
        if not resp.ok:
            raise ValueError(resp.error_message())
        if verbose:
            if resp.empty():
                logger.info("Graph is configured correctly.")
            else:
                logger.warning(resp.message())
        return self

    def visualize(
        self,
        path: Optional[Union[str, io.BytesIO]] = None,
        show_cols: bool = True,
    ) -> 'graphviz.Graph':
        r"""Visualizes the tables and edges in this graph using the
        ``graphviz`` library.

        Args:
            path: An optional local path to write the produced image to. If
                None, the image will not be written to disk.
            show_cols: Whether to show all columns of every table in the graph.
                If False, will only show the primary key, foreign key(s),
                time column, and end time column of each table.

        Returns:
            A ``graphviz.Graph`` instance representing the visualized graph.
        """
        def has_graphviz_executables() -> bool:
            import graphviz
            try:
                graphviz.Digraph().pipe()
            except graphviz.backend.ExecutableNotFound:
                return False

            return True

        # Check basic dependency:
        if not find_spec('graphviz'):
            raise ModuleNotFoundError(
                "The `graphviz` Python package is required for visualization.")
        elif not has_graphviz_executables():
            raise RuntimeError(
                "Could not visualize graph as `graphviz` executables have not "
                "been installed. These dependencies are required in addition "
                "to the `graphviz` Python package. Please install them to "
                "continue. Instructions at https://graphviz.org/download/.")
        else:
            import graphviz

        fmt = None
        if isinstance(path, str):
            fmt = path.split('.')[-1]
        elif isinstance(path, io.BytesIO):
            fmt = 'svg'
        graph = graphviz.Graph(format=fmt)

        def left_align(list_of_text: List[str]) -> str:
            return '\\l'.join(list_of_text) + '\\l'

        table_to_fkey: Dict[str, List[str]] = {}
        for edge in self.edges:
            src, fkey, dst = edge
            if src not in table_to_fkey:
                table_to_fkey[src] = []
            table_to_fkey[src].append(fkey)

        for table_name, table in self.tables.items():
            keys = []
            if table.has_primary_key():
                assert table.primary_key is not None
                keys += [f'{table.primary_key.name} (PK)']
            if table_name in table_to_fkey:
                keys += [f'{fkey} (FK)' for fkey in table_to_fkey[table_name]]
            if table.has_time_column():
                assert table.time_column is not None
                keys += [f'{table.time_column.name} (Time)']
            if table.has_end_time_column():
                assert table.end_time_column is not None
                keys += [f'{table.end_time_column.name} (End Time)']

            keys_aligned = left_align(keys)

            cols = []
            cols_aligned = ""
            if show_cols and len(table.columns) > 0:
                cols += [
                    f'{col.name}: {col.stype or "???"} ({col.dtype or "???"})'
                    for col in table.columns
                ]
                cols_aligned = left_align(cols)

            if cols:
                label = f'{{{table_name}|{keys_aligned}|{cols_aligned}}}'
            else:
                label = f'{{{table_name}|{keys_aligned}}}'

            graph.node(table_name, shape='record', label=label)

        for edge in self.edges:
            src, fkey, dst = edge
            pkey_obj = self[dst].primary_key
            assert pkey_obj is not None
            pkey = pkey_obj.name
            # Print both key names if different:
            if fkey != pkey:
                label = f' {fkey}\n< >\n{pkey} '
            else:
                label = f' {fkey} '
            headlabel, taillabel = '1', '*'
            graph.edge(src, dst, label=label, headlabel=headlabel,
                       taillabel=taillabel, minlen='2', fontsize='11pt',
                       labeldistance='1.5')

        if isinstance(path, str):
            path = '.'.join(path.split('.')[:-1])
            graph.render(path, cleanup=True)
        elif isinstance(path, io.BytesIO):
            path.write(graph.pipe())
        else:
            try:
                graph.view()
            except Exception as e:
                logger.warning(
                    "Could not visualize graph due to an unexpected error in "
                    "`graphviz`. If you are in a notebook environment, "
                    "consider calling `display()` on the returned object "
                    "from `visualize()`. Error: %s", e)
        return graph

    # Class properties ########################################################

    def __hash__(self) -> int:
        return hash((tuple(self.edges), self.tables.values()))

    def __contains__(self, name: str) -> bool:
        return self.has_table(name)

    def __getitem__(self, name: str) -> Table:
        return self.table(name)

    def __delitem__(self, name: str) -> None:
        self.remove_table(name)

    def __repr__(self) -> str:
        table_names = str(list(self._tables.keys())).replace("'", "")
        return (f'{self.__class__.__name__}(\n'
                f'  tables={table_names},\n'
                f'  edges={self._edges},\n'
                f')')



from abc import ABC, abstractmethod
from typing import List, Set, Union

from kumoapi.data_source import DataSourceType
from kumoapi.source_table import (
    BigQuerySourceTableRequest,
    DatabricksSourceTableRequest,
    S3SourceTableRequest,
    SnowflakeSourceTableRequest,
    SourceTableConfigRequest,
    SourceTableConfigResponse,
    SourceTableDataRequest,
    SourceTableDataResponse,
    SourceTableListRequest,
    SourceTableListResponse,
    SourceTableValidateRequest,
    SourceTableValidateResponse,
)

from kumoai import global_state
from kumoai.connector.source_table import SourceTable
from kumoai.exceptions import HTTPException


class Connector(ABC):
    r"""A connector to a backing data source, that can be used to create Kumo
    tables.
    """

    _validated_tables: Set[str] = set()

    # Metadata ################################################################

    @property
    @abstractmethod
    def source_type(self) -> DataSourceType:
        r"""Returns the data source type accessible by this connector."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self) -> str:
        r"""Returns the name of the connector.

        .. note::
            If the connector does not support naming, the name refers to an
            internal specifier.
        """
        raise NotImplementedError()

    @abstractmethod
    def _source_table_request(
        self,
        table_names: List[str],
    ) -> Union[S3SourceTableRequest, BigQuerySourceTableRequest,
               DatabricksSourceTableRequest, SnowflakeSourceTableRequest]:
        raise NotImplementedError()

    # Tables ##################################################################

    def table_names(self) -> List[str]:
        r"""Returns a list of table names accessible through this connector."""
        return self._list_tables().table_names

    def has_table(self, name: str) -> bool:
        r"""Returns :obj:`True` if the table exists in this connector,
        :obj:`False` otherwise.

        Args:
            name: The table name.
        """
        try:
            resp = self._validate_table(name)
            return resp.is_valid
        except HTTPException:
            # In case of HTTPException, Kumo backend doesn't have api
            # implemented and we skip check by returns True.
            return True

    def table(self, name: str) -> SourceTable:
        r"""Returns a :class:`~kumoai.connector.SourceTable` object
        corresponding to a source table behind this connector. A source table
        is a view into the raw data of table :obj:`name`. To use a source
        table in Kumo, you will need to construct a
        :class:`~kumoai.graph.Table` from the source table.

        Args:
            name: The table name.

        Raises:
            :class:`ValueError`: if :obj:`name` does not exist in the backing
                connector.
        """
        if not self.has_table(name):
            raise ValueError(f"The table '{name}' does not exist in {self}. "
                             f"Please check the existence of the source data.")

        return SourceTable(name=name, connector=self)

    def _validate_table(self, table_name: str) -> SourceTableValidateResponse:
        if table_name in self._validated_tables:
            return SourceTableValidateResponse(is_valid=True, msg='')

        req = SourceTableValidateRequest(
            connector_id=self.name,
            table_name=table_name,
            source_type=self.source_type,
        )
        ret = global_state.client.source_table_api.validate_table(req)

        # Cache the result for the whole session.
        if ret.is_valid:
            self._validated_tables.add(table_name)
        return ret

    def _list_tables(self) -> SourceTableListResponse:
        req = SourceTableListRequest(connector_id=self.name,
                                     source_type=self.source_type)
        return global_state.client.source_table_api.list_tables(req)

    def _get_table_data(
        self,
        table_names: List[str],
        sample_rows: int,
    ) -> SourceTableDataResponse:
        req = SourceTableDataRequest(
            source_table_request=self._source_table_request(table_names),
            sample_rows=sample_rows,
        )
        return global_state.client.source_table_api.get_table_data(req)

    def _get_table_config(self, table_name: str) -> SourceTableConfigResponse:
        # TODO(manan): rest backend for this is a bit broken, it never returns
        # directories...
        req = SourceTableConfigRequest(connector_id=self.name,
                                       table_name=table_name,
                                       source_type=self.source_type)
        return global_state.client.source_table_api.get_table_config(req)

    # Class properties ########################################################

    def __hash__(self) -> int:
        return hash(self.__dict__)

    def __contains__(self, name: str) -> bool:
        return self.has_table(name)

    def __getitem__(self, name: str) -> SourceTable:
        return self.table(name)

    def __repr__(self) -> str:
        # TODO(manan): class-overrideable metadata?
        return f'{self.__class__.__name__}()'



import copy
from dataclasses import fields, make_dataclass
from http import HTTPStatus
from typing import List, Optional, Tuple, Union, get_args, get_origin

from kumoapi.common import ValidationResponse
from kumoapi.json_serde import to_json_dict
from kumoapi.model_plan import (
    MissingType,
    ModelPlan,
    ModelPlanInfo,
    PlanMixin,
    SuggestModelPlanRequest,
    TrainingTableGenerationPlan,
)
from kumoapi.pquery import PQueryResource, QueryType
from kumoapi.task import TaskType
from pydantic.dataclasses import dataclass

from kumoai.client import KumoClient
from kumoai.client.endpoints import PQueryEndpoints
from kumoai.client.utils import (
    Returns,
    parse_id_response,
    parse_response,
    raise_on_error,
)

GraphID = str
PQueryID = str


class PQueryAPI:
    r"""Typed API definition for Kumo Predictive Query definition."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client

    def create(self, pquery: PQueryResource) -> PQueryID:
        r"""Creates a Predictive Query resource object in Kumo."""
        resp = self._client._request(PQueryEndpoints.create,
                                     json=to_json_dict(pquery))
        raise_on_error(resp)
        return parse_id_response(resp)

    def get_if_exists(self, name_or_id: str) -> Optional[PQueryResource]:
        r"""Fetches a Predictive Query given its name(id)."""
        resp = self._client._request(PQueryEndpoints.get.with_id(name_or_id))
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise_on_error(resp)
        return parse_response(PQueryResource, resp)

    def list(self, *,
             name_pattern: Optional[str] = None) -> List[PQueryResource]:
        params = {}
        if name_pattern:
            params['name_pattern'] = name_pattern
        resp = self._client._request(PQueryEndpoints.list, params=params)
        raise_on_error(resp)
        return parse_response(List[PQueryResource], resp)

    def delete(self, name: str) -> None:
        resp = self._client._request(PQueryEndpoints.delete.with_id(name))
        raise_on_error(resp)

    def infer_task_type(
        self,
        pquery_string: str,
        graph_id: GraphID,
    ) -> Tuple[TaskType, str]:
        resp = self._client._request(
            PQueryEndpoints.infer_task_type,
            params={
                'pquery_string': pquery_string,
                'graph_id': graph_id
            },
        )
        raise_on_error(resp)
        return parse_response(Returns[Tuple[TaskType, str]], resp)

    def validate(self, pquery: PQueryResource) -> ValidationResponse:
        r"""Validates a Predictive Query resource object in Kumo."""
        resp = self._client._request(PQueryEndpoints.validate,
                                     json=to_json_dict(pquery))
        raise_on_error(resp)
        return parse_response(Returns[ValidationResponse], resp)

    def suggest_training_table_plan(
        self,
        req: SuggestModelPlanRequest,
    ) -> TrainingTableGenerationPlan:
        resp = self._client._request(
            PQueryEndpoints.suggest_training_table_plan,
            json=to_json_dict(req),
        )
        raise_on_error(resp)
        return parse_response(TrainingTableGenerationPlan, resp)

    def suggest_model_plan(
        self,
        req: SuggestModelPlanRequest,
    ) -> ModelPlan:
        resp = self._client._request(
            PQueryEndpoints.suggest_model_plan,
            json=to_json_dict(req),
        )
        raise_on_error(resp)
        plan_info = parse_response(ModelPlanInfo, resp)

        return filter_model_plan(
            plan_info.model_plan,
            plan_info.task_type,
            plan_info.query_type,
            plan_info.has_train_table_weight_col,
        )


def filter_model_plan(
    plan: ModelPlan,
    task_type: TaskType,
    query_type: QueryType,
    has_train_table_weight_col: bool,
) -> ModelPlan:
    r"""Filters the model plan ``plan`` down to its valid fields only, given
    ``task_type`` and ``query_type``.
    """
    new_section_fields = []
    new_sections = []
    for section_field in fields(plan):
        section = getattr(plan, section_field.name)

        new_opt_fields = []
        new_opts = []
        for field in fields(section):
            metadata = field.default.extra['metadata']  # type: ignore
            if not section.is_valid_option(
                    name=field.name,
                    metadata=metadata,
                    task_type=task_type,
                    query_type=query_type,
                    has_train_table_weight_col=has_train_table_weight_col,
            ):
                continue

            # Remove `MissingType` from type annotation:
            if not metadata.hidden and get_origin(field.type) is Union:
                _types = tuple(_type for _type in get_args(field.type)
                               if _type is not MissingType)
                assert len(_types) > 0
                _type = Union[_types] if len(_types) > 1 else _types[0]
            else:
                _type = field.type

            default = copy.copy(field.default)
            try:
                # Pydantic v1
                from pydantic.fields import Undefined  # type: ignore
                default.default = Undefined  # type: ignore
            except ImportError:
                try:
                    # Pydantic v2 - Undefined moved to pydantic_core
                    from pydantic_core import PydanticUndefined as Undefined
                    default.default = Undefined  # type: ignore
                except ImportError:
                    # Fallback: Neither v1 nor v2 Undefined available
                    # In Pydantic v2, not setting a default is equivalent to
                    # Undefined
                    pass

            new_opt_fields.append((field.name, _type, default))
            new_opts.append(getattr(section, field.name))

        Section = make_dataclass(
            section_field.type.__name__,  # type: ignore
            new_opt_fields,
            bases=(PlanMixin, ),
            repr=False,
        )
        Section = dataclass(
            config=dict(validate_assignment=True),
            repr=False,
        )(Section)
        new_section_fields.append((section_field.name, Section))
        new_sections.append(Section(*new_opts))

    Plan = make_dataclass(
        'ModelPlan',
        new_section_fields,
        bases=(ModelPlan, ),
        repr=False,
    )
    Plan = dataclass(
        config=dict(validate_assignment=True),
        repr=False,
    )(Plan)
    return Plan(*new_sections)



from dataclasses import dataclass
from typing import Any, Optional, Union

from kumoapi.table import TimestampUnit
from kumoapi.typing import Dtype, Stype

from kumoai.mixin import CastMixin


@dataclass(init=False)
class Column(CastMixin):
    r"""A column represents metadata information for a column in a Kumo
    :class:`~kumoai.graph.Table`. Columns can be created independent of
    a table, or can be fetched from a table with the
    :meth:`~kumoai.graph.Table.column` method.

    .. code-block:: python

        import kumoai

        # Fetch a column from a `kumoai.Table`:
        table = kumoai.Table(...)

        column = table.column("col_name")
        column = table["col_name"]  # equivalent to the above.

        # Edit a column's data type:
        print("Existing dtype: ", column.dtype)
        column.dtype = "int"

        # Edit a column's semantic type:
        print("Existing stype: ", column.stype)
        column.stype = "ID"

    Args:
        name: The name of this column.
        stype: The semantic type of this column. Semantic types can be
            specified as strings: the list of possible semantic types
            is located at :class:`~kumoai.Stype`.
        dtype: The data type of this column. Data types can be specified
            as strings: the list of possible data types is located at
            :class:`~kumoai.Dtype`.
        timestamp_format: If this column represents a timestamp, the format
            that the timestamp should be parsed in. The format can either be
            a :class:`~kumoapi.table.TimestampUnit` for integer columns or a
            string with a format identifier described
            `here <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`__
            for a SaaS Kumo deployment and
            `here <https://docs.snowflake.com/en/sql-reference/date-time-input-output#about-the-elements-used-in-input-and-output-formats>`__
            for a Snowpark Container Services Kumo deployment. If left empty,
            will be intelligently inferred by Kumo.
    """  # noqa: E501
    name: str
    stype: Optional[Stype] = None
    dtype: Optional[Dtype] = None
    timestamp_format: Optional[Union[str, TimestampUnit]] = None

    def __init__(
        self,
        name: str,
        stype: Optional[Union[Stype, str]] = None,
        dtype: Optional[Union[Dtype, str]] = None,
        timestamp_format: Optional[Union[str, TimestampUnit]] = None,
    ) -> None:
        self.name = name
        self.stype = Stype(stype) if stype is not None else None
        self.dtype = Dtype(dtype) if dtype is not None else None
        try:
            self.timestamp_format = TimestampUnit(timestamp_format)
        except ValueError:
            self.timestamp_format = timestamp_format

    def __hash__(self) -> int:
        return hash((self.name, self.stype, self.dtype, self.timestamp_format))

    def __setattr__(self, key: Any, value: Any) -> None:
        if key == 'name' and value != getattr(self, key, value):
            raise AttributeError("Attribute 'name' is read-only")
        elif key == 'stype' and isinstance(value, str):
            value = Stype(value)
        elif key == 'dtype' and isinstance(value, str):
            value = Dtype(value)
        elif key == 'timestamp_format' and isinstance(value, str):
            try:
                value = TimestampUnit(value)
            except ValueError:
                pass
        super().__setattr__(key, value)

    def update(self, obj: 'Column', override: bool = True) -> 'Column':
        for key in self.__dict__:
            if key[0] == '_':  # Skip private attributes:
                continue
            value = getattr(obj, key, None)
            if value is not None:
                if override or getattr(self, key, None) is None:
                    setattr(self, key, value)
        return self

    def __repr__(self) -> str:
        out = (f"Column(name=\"{self.name}\", stype=\"{self.stype}\", "
               f"dtype=\"{self.dtype}\"")
        if self.timestamp_format is not None:
            out += f", timestamp_format=\"{self.timestamp_format}\""
        out += ")"
        return out



import asyncio
import os
from functools import reduce
from typing import TYPE_CHECKING, Dict, List, Optional

if TYPE_CHECKING:
    from snowflake.snowpark import DataFrame, Session


def _get_spcs_token(snowflake_credentials: Dict[str, str]) -> str:
    r"""Fetches a token to access a Kumo application deployed in Snowflake
    Snowpark Container Services (SPCS). This token is valid for 1 hour, after
    which the token must be re-generated.
    """
    # Create a request to the ingress endpoint with authz:
    active_session = _get_active_session()
    if active_session is not None:
        ctx = active_session.connection
    else:
        user = snowflake_credentials["user"]
        password = snowflake_credentials["password"]
        account = snowflake_credentials["account"]
        import snowflake.connector
        ctx = snowflake.connector.connect(
            user=user,
            password=password,
            account=account,
            session_parameters={
                'PYTHON_CONNECTOR_QUERY_RESULT_FORMAT': 'json'
            },
        )

    # Obtain a session token:
    token_data = ctx._rest._token_request('ISSUE')
    token_extract = token_data['data']['sessionToken']
    return f'\"{token_extract}\"'


def _refresh_spcs_token() -> None:
    r"""Refreshes the SPCS token in global state to avoid expiration."""
    from kumoai import KumoClient, global_state
    if (not global_state.initialized
            or (not global_state._snowflake_credentials
                and not global_state._snowpark_session)):
        raise ValueError(
            "Please initialize the Kumo application with snowflake "
            "credentials before attempting to refresh this token.")
    global_state._spcs_token = _get_spcs_token(
        global_state._snowflake_credentials or {})

    # Verify token validity:
    assert global_state._url is not None
    client = KumoClient(
        url=global_state._url,
        api_key=global_state._api_key,
        spcs_token=global_state._spcs_token,
    )
    if not client.authenticate():
        raise ValueError("Client authentication failed. Please check if you "
                         "have a valid API key.")

    # Update state:
    global_state._spcs_token = client._spcs_token


async def _run_refresh_spcs_token(minutes: int) -> None:
    r"""Runs the SPCS token refresh loop every `minutes` minutes."""
    while True:
        await asyncio.sleep(minutes * 60)
        _refresh_spcs_token()


def _get_active_session() -> 'Optional[Session]':
    try:
        from snowflake.snowpark.context import get_active_session
        return get_active_session()
    except Exception:
        return None


def _get_session() -> 'Session':
    import snowflake.snowpark as snowpark

    from kumoai import global_state
    params = global_state._snowflake_credentials
    assert params is not None

    database = os.getenv("SNOWFLAKE_DATABASE")
    schema = os.getenv("SNOWFLAKE_SCHEMA")
    if not database or not schema:
        raise ValueError("Please set the SNOWFLAKE_DATABASE and "
                         "SNOWFLAKE_SCHEMA environment variables.")
    params['database'] = database
    params['schema'] = schema
    params['client_session_keep_alive'] = True

    return snowpark.Session.builder.configs(params).create()


def _remove_path(session: 'Session', stage_path: str, file_path: str) -> None:
    stage_prefix = '.'.join(stage_path.split('.')[:2])
    name_remove = '.'.join([stage_prefix, file_path])
    session.sql(f"REMOVE {name_remove}").collect()


def _parquet_to_df(path: str) -> 'DataFrame':
    r"""Reads parquet from the given path and returns a snowpark DataFrame."""
    session = _get_session()
    if not path.endswith(os.path.sep):
        path += os.path.sep
    file_list = session.sql(f"LIST {path}").collect()
    for file_row in file_list:
        if file_row.name.endswith('.parquet'):
            continue
        _remove_path(session, path, file_row.name)
    df = session.read.parquet(path)
    return df


def _parquet_dataset_to_df(paths: List[str]) -> 'DataFrame':
    r"""Reads parquet from the given paths and returns a snowpark DataFrame."""
    from snowflake.snowpark import DataFrame
    df_list = [_parquet_to_df(url) for url in paths]
    return reduce(DataFrame.union_all, df_list)



import asyncio
import concurrent
import concurrent.futures
import time

from kumoapi.common import JobStatus
from typing_extensions import override

from kumoai import global_state
from kumoai.futures import KumoProgressFuture, create_future


class ArtifactExportResult:
    r"""Represents a completed artifact export job."""
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id

    def tracking_url(self) -> str:
        r"""Returns a tracking URL pointing to the UI display of
        this prediction export job.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(job_id={self.job_id})"


class ArtifactExportJob(KumoProgressFuture[ArtifactExportResult]):
    """Represents an in-progress artifact export job."""
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id
        self._fut: concurrent.futures.Future = create_future(
            _poll_export(job_id))

    @property
    def id(self) -> str:
        """The unique ID of this export job."""
        return self.job_id

    @override
    def result(self) -> ArtifactExportResult:
        return self._fut.result()

    @override
    def future(self) -> 'concurrent.futures.Future[ArtifactExportResult]':
        return self._fut

    @override
    def _attach_internal(
        self,
        interval_s: float = 20.0,
    ) -> ArtifactExportResult:
        """Allows a user to attach to a running export job and view
        its progress.

        Args:
            interval_s (float): Time interval (seconds) between polls, minimum
                value allowed is 4 seconds.
        """
        assert interval_s >= 4.0
        print(f"Attaching to export job {self.job_id}. To detach from "
              f"this job, please enter Ctrl+C (the job will continue to run, "
              f"and you can re-attach anytime).")

        # TODO improve print statements.
        # Will require changes to status to return
        # JobStatusReport instead of JobStatus.
        while not self.done():
            status = self.status()
            print(f"Export job {self.job_id} status: {status}")
            time.sleep(interval_s)

        return self.result()

    def status(self) -> JobStatus:
        """Returns the status of a running export job."""
        return get_export_status(self.job_id)

    def cancel(self) -> bool:
        """Cancels a running export job.

        Returns:
            bool: True if the job is in a terminal state.
        """
        api = global_state.client.artifact_export_api
        status = api.cancel(self.job_id)
        if status == JobStatus.CANCELLED:
            return True
        return False


def get_export_status(job_id: str) -> JobStatus:
    api = global_state.client.artifact_export_api
    resource = api.get(job_id)
    return resource


async def _poll_export(job_id: str) -> ArtifactExportResult:
    status = get_export_status(job_id)
    while not status.is_terminal:
        await asyncio.sleep(10)
        status = get_export_status(job_id)

    if status != JobStatus.DONE:
        raise RuntimeError(f"Export job {job_id} failed "
                           f"with job status {status}.")

    return ArtifactExportResult(job_id=job_id)



from kumoai.connector import FileUploadConnector
from kumoai.connector.utils import replace_table
from kumoai.graph import Edge, Graph, Table


def from_relbench(dataset_name: str) -> Graph:
    r"""Creates a Kumo graph from a RelBench dataset. This function processes
    the specified RelBench dataset, uploads its tables to the Kumo data plane,
    and registers them as part of a Kumo graph, including inferred table
    metadata and edges.

    .. note::

        Please note that this method is subject to the limitations for file
        upload in :class:`~kumoai.connector.FileUploadConnector`.

    .. code-block:: python

        import kumoai
        from kumoai.utils.datasets import from_relbench

        # Assume dataset `example_dataset` in the RelBench repository:
        graph = from_relbench(dataset_name="example_dataset")

    Args:
        dataset_name: The name of the RelBench dataset to be processed.

    Returns:
        A :class:`~kumoai.Graph` object containing the tables and edges
        derived from the RelBench dataset.

    Raises:
        ValueError: If the dataset cannot be retrieved or processed.
    """
    try:
        import relbench
    except ImportError:
        raise RuntimeError(
            "Creating a Kumo Graph from a RelBench dataset requires the "
            "'relbench' package to be installed. Please install the package "
            "before proceeding.")

    connector = FileUploadConnector(file_type="parquet")
    dataset = relbench.datasets.get_dataset(dataset_name, download=True)
    db = dataset.get_db(upto_test_timestamp=False)

    # Store table metadata and edges:
    table_metadata = {}

    # Process each table in the database
    for table_key in db.table_dict.keys():
        # Save the table locally as a parquet file:
        table = db.table_dict[table_key]
        parquet_path = f"tmp_{table_key}.parquet"
        table.df.to_parquet(parquet_path, index=False)

        # Replace the table on the Kumo data plane:
        replace_table(name=table_key, path=parquet_path, file_type="parquet")

        # Register the table with inferred metadata and collect edge
        # information:
        table_metadata[table_key] = dict(
            table=Table.from_source_table(
                source_table=connector[table_key],
                primary_key=table.pkey_col,
                time_column=table.time_col,
            ).infer_metadata(), edges=table.fkey_col_to_pkey_table)

    tables = {
        table_key: table_metadata[table_key]['table']
        for table_key in table_metadata.keys()
    }
    edges = []
    for table_key, table_data in table_metadata.items():
        for edge_key, dst_table in table_data['edges'].items():
            edges.append(
                Edge(src_table=table_key, fkey=edge_key, dst_table=dst_table))

    # Create and return the graph
    return Graph(
        tables=tables,
        edges=edges,
    )



from .trainer import Trainer
from kumoapi.model_plan import (
    TrainingJobPlan,
    ColumnProcessingPlan,
    NeighborSamplingPlan,
    OptimizationPlan,
    ModelArchitecturePlan,
    ModelPlan,
    GNNModelPlan,
    GraphTransformerModelPlan,
)
# For backwards compatibility
from kumoai.artifact_export import (
    ArtifactExportJob,
    ArtifactExportResult,
)
from .job import (
    TrainingJobResult,
    TrainingJob,
    BatchPredictionJobResult,
    BatchPredictionJob,
)
from .baseline_trainer import BaselineTrainer

__all__ = [
    'TrainingJobPlan',
    'ColumnProcessingPlan',
    'NeighborSamplingPlan',
    'OptimizationPlan',
    'ModelArchitecturePlan',
    'ModelPlan',
    'GNNModelPlan',
    'GraphTransformerModelPlan',
    'Trainer',
    'TrainingJobResult',
    'TrainingJob',
    'BatchPredictionJobResult',
    'BatchPredictionJob',
    'BaselineTrainer',
    'ArtifactExportJob',
    'ArtifactExportResult',
]



import os
from typing import Dict, List, Optional

from kumoapi.data_source import (
    CreateConnectorArgs,
    DatabricksConnectorResourceConfig,
    DatabricksCredentials,
    DataSourceType,
)
from kumoapi.source_table import DatabricksSourceTableRequest
from typing_extensions import Self, override

from kumoai import global_state
from kumoai.connector import Connector

_ENV_DATABRICKS_CLIENT_ID = 'DATABRICKS_CLIENT_ID'
_ENV_DATABRICKS_CLIENT_SECRET = 'DATABRICKS_CLIENT_SECRET'
_ENV_DATABRICKS_TOKEN = 'DATABRICKS_TOKEN'


class DatabricksConnector(Connector):
    r"""Establishes a connection to a
    `Databricks <https://www.databricks.com/>`_ database.

    Authentication requires passing either a client ID and client secret, or a
    personal access token, to the connector, either via environment variables
    (``DATABRICKS_CLIENT_ID`` and ``DATABRICKS_CLIENT_SECRET``, or
    ``DATABRICKS_TOKEN``), or via keys in the credentials dictionary
    (``client_id`` and ``client_secret``, or ``token``).

    .. code-block:: python

        import kumoai

        # Either pass `credentials=dict(client_id=..., client_secret=...,
        # token=...) or set the 'DATABRICKS_CLIENT_ID' and
        # 'DATABRICKS_CLIENT_SECRET' (or 'DATABRICKS_TOKEN') environment
        # variables:
        connector = kumoai.connector.DatabricksConnector(
            name="<connector_name>",
            host="<databricks_host_name>",
            cluster_id="<databricks_cluster_id>",
            warehouse_id="<databricks_warehouse_id>",
            catalog="<databricks_catalog_name>",
            credentials=credentials,
        )

        # List all tables:
        print(connector.table_names())

        # Check whether a table is present:
        assert "articles" in connector

        # Fetch a source table (both approaches are equivalent):
        source_table = connector["articles"]
        source_table = connector.table("articles")

    Args:
        name: The name of the connector.
        host: The host name.
        cluster_id: The cluster ID of this warehouse.
        warehouse_id: The warehouse ID of this warehous.
        catalog: The name of the Databricks catalog.
        credentials: The client ID, client secret, and personal access token
            that correspond to this Databricks account.
    """
    def __init__(
            self,
            name: str,
            host: str,
            cluster_id: str,
            warehouse_id: str,
            catalog: str,
            credentials: Optional[Dict[str, str]] = None,
            _bypass_creation: bool = False,  # INTERNAL ONLY.
    ):
        super().__init__()

        self._name = name
        self.host = host
        self.cluster_id = cluster_id
        self.warehouse_id = warehouse_id
        self.catalog = catalog

        if _bypass_creation:
            # TODO(manan, siyang): validate that this connector actually exists
            # in the REST DB:
            return

        # Fully specify credentials, create Kumo connector:
        credentials = credentials or {}
        credentials_args = {
            "client_id":
            credentials.get("client_id", os.getenv(_ENV_DATABRICKS_CLIENT_ID)),
            "client_secret":
            credentials.get("client_secret",
                            os.getenv(_ENV_DATABRICKS_CLIENT_SECRET)),
            "token":
            credentials.get("token", os.getenv(_ENV_DATABRICKS_TOKEN))
        }

        has_pat = credentials_args["token"] is not None
        has_client_id_secret = (credentials_args["client_id"] is not None and
                                credentials_args["client_secret"] is not None)

        if has_pat and has_client_id_secret:
            raise ValueError(
                "Please pass only one of a (Databricks client ID and "
                "Databricks client secret) or a (Databricks PAT).")
        elif not (has_pat or has_client_id_secret):
            raise ValueError(
                f"Please pass valid credentials to create a Databricks "
                f"connector. You can do so either via the 'credentials' "
                f"argument or the {_ENV_DATABRICKS_CLIENT_ID} and "
                f"{_ENV_DATABRICKS_CLIENT_SECRET}, or "
                f"{_ENV_DATABRICKS_TOKEN} environment variables.")

        self._create_connector(credentials_args)  # type: ignore

    @classmethod
    def get_by_name(cls, name: str) -> Self:
        r"""Returns an instance of a named Databricks Connector, including
        those created in the Kumo UI.

        Args:
            name: The name of the existing connector.

        Example:
            >>> import kumoai
            >>> connector = kumoai.DatabricksConnector.get_by_name("name")  # doctest: +SKIP # noqa: E501
        """
        api = global_state.client.connector_api
        resp = api.get(name)
        if resp is None:
            raise ValueError(
                f"There does not exist an existing stored connector with name "
                f"{name}.")
        config = resp.config
        assert isinstance(config, DatabricksConnectorResourceConfig)
        return cls(
            name=config.name,
            host=config.host,
            cluster_id=config.cluster_id,
            warehouse_id=config.warehouse_id,
            catalog=config.catalog,
            credentials=None,
            _bypass_creation=True,
        )

    @override
    @property
    def name(self) -> str:
        r"""Returns the name of this connector."""
        return self._name

    @override
    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.DATABRICKS

    @override
    def _source_table_request(
        self,
        table_names: List[str],
    ) -> DatabricksSourceTableRequest:
        return DatabricksSourceTableRequest(
            connector_id=self.name,
            table_names=table_names,
        )

    def _create_connector(self, credentials: Dict[str, str]) -> None:
        r"""Creates and persists a Databricks connector in the REST DB.
        Currently only intended for internal use.

        Args:
            credentials: Fully-specified credentials containing the username
                and password for the Databricks connector.

        Raises:
            RuntimeError: if connector creation failed
        """
        credentials = DatabricksCredentials(
            client_id=credentials["client_id"] or '',
            client_secret=credentials["client_secret"] or '',
            pat=credentials["token"] or '',
        )
        args = CreateConnectorArgs(
            config=DatabricksConnectorResourceConfig(
                name=self.name,
                host=self.host,
                cluster_id=self.cluster_id,
                warehouse_id=self.warehouse_id,
                catalog=self.catalog,
            ),
            credentials=credentials,
        )
        global_state.client.connector_api.create_if_not_exist(args)

    def _delete_connector(self) -> None:
        r"""Deletes a connector in the REST DB. Only intended for internal
        use.
        """
        global_state.client.connector_api.delete_if_exists(self.name)

    # Class properties ########################################################

    @override
    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}'
                f'(name=\"{self.name}\", host=\"{self.host}\", '
                f'cluster_id=\"{self.cluster_id}\", '
                f'warehouse_id=\"{self.warehouse_id}\", '
                f'catalog=\"{self.catalog}\")')



import os
import tempfile
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Set, Union

import pandas as pd

from kumoai.connector import FileUploadConnector
from kumoai.connector.utils import upload_table
from kumoai.experimental.rfm import LocalTable
from kumoai.experimental.rfm.utils import dtype_to_family
from kumoai.graph import Edge, Graph, Table

MAX_TABLE_SIZE_MB = 10 * 1024  # 10GB in MB


class LocalGraph:
    r"""A graph of :class:`LocalTable` objects, akin to relationships between
    tables in a relational database.

    Creating a graph is the final step of data definition; after a
    :class:`LocalGraph` is created, you can use it to initialize the
    Kumo Relational Foundation Model (:class:`KumoRFM`).

    .. code-block:: python

        import kumoai.experimental.rfm as rfm

        # dataframes
        df1 = pd.DataFrame(...)
        df2 = pd.DataFrame(...)
        df3 = pd.DataFrame(...)

        # define tables
        table1 = kumoai.LocalTable(name="table1", data=df1)
        table2 = kumoai.LocalTable(name="table2", data=df2)
        table3 = kumoai.LocalTable(name="table3", data=df3)

        # create a graph from a list of tables
        graph = kumoai.LocalGraph(
            tables={
                "table1": table1,
                "table2": table2,
                "table3": table3,
            },
            edges=[],
        )

        # infer links
        graph.infer_links()

        # remove edges between tables
        graph.unlink(table1, table2, fkey="id1")

        # infer metadata
        graph.infer_metadata()

        # validate graph
        graph.validate()

        # construct a graph from dataframes
        graph = rfm.LocalGraph.from_data(data={
            "table1": df1,
            "table2": df2,
            "table3": df3,
        })

        # remove edge between tables
        graph.unlink(table1, table2, fkey="id1")

        # validate graph
        graph.validate()

        # re-link tables
        graph.link(table1, table2, fkey="id1")
    """

    # Constructors ############################################################

    def __init__(
        self,
        tables: List[LocalTable],
        edges: Optional[List[Edge]] = None,
    ) -> None:
        self._tables: Dict[str, LocalTable] = {}
        self._edges: List[Edge] = []

        if len(tables) != len(set([table.table_name for table in tables])):
            raise ValueError("Table names must be globally unique within a "
                             "graph.")

        for table in tables:
            self._tables[table.table_name] = table

        for edge in (edges or []):
            self.link(Edge._cast(edge))

    @staticmethod
    def from_data(
        data: Dict[str, pd.DataFrame],
        edges: Optional[List[Edge]] = None,
    ) -> 'LocalGraph':
        r"""Creates a :class:`LocalGraph` from a dictionary of
        :class:`pandas.DataFrame` objects.

        Args:
            data: A dictionary of data frames, where the keys are the names of
                the tables and the values hold table data.
            edges: An optional list of :class:`~kumoai.graph.Edge` objects to
                add to the graph. If not provided, edges will be automatically
                inferred from the data.

        Note:
            This method will automatically infer metadata and links for the
            graph.

        Example:
            >>> import kumoai.experimental.rfm as rfm
            >>> df1 = pd.DataFrame(...)
            >>> df2 = pd.DataFrame(...)
            >>> df3 = pd.DataFrame(...)
            >>> graph = rfm.LocalGraph.from_data(data={
            ...     "table1": df1,
            ...     "table2": df2,
            ...     "table3": df3,
            ... })
            ... graph.validate()  # doctest: +SKIP
        """
        tables = [
            LocalTable(df=df, table_name=name) for name, df in data.items()
        ]

        graph = LocalGraph(tables=tables, edges=edges or []).infer_metadata()

        if edges is None:
            graph.infer_links()

        return graph

    # Tables ##############################################################

    def has_table(self, name: str) -> bool:
        r"""Returns ``True`` if this graph has a table with name
        ``name``; ``False`` otherwise.
        """
        return name in self.tables

    def table(self, name: str) -> LocalTable:
        r"""Returns the table with name ``name`` in this graph.

        Raises:
            KeyError: If ``name`` is not present in this graph.

        """
        if not self.has_table(name):
            raise KeyError(f"Table '{name}' not found in graph")
        return self.tables[name]

    @property
    def tables(self) -> Dict[str, LocalTable]:
        r"""Returns the dictionary of table objects."""
        return self._tables

    def infer_metadata(self) -> 'LocalGraph':
        r"""Infers metadata for the tables in this :class:`LocalGraph`, by
        inferring the metadata of each :class:`LocalTable` in the graph.

        Note:
            For more information, please see
            :meth:`kumoai.experimental.rfm.LocalTable.infer_metadata`.
        """
        for table in self.tables.values():
            table.infer_metadata(inplace=True)

        return self

    # Edges ###################################################################

    @property
    def edges(self) -> List[Edge]:
        r"""Returns the edges of this graph."""
        return self._edges

    def link(
        self,
        *args: Optional[Union[str, Edge]],
        **kwargs: str,
    ) -> 'LocalGraph':
        r"""Links two tables (``src_table`` and ``dst_table``) from the foreign
        key ``fkey`` in the source table to the primary key in the destination
        table.

        These edges are treated as bidirectional.

        Args:
            *args: Any arguments to construct an :class:`~kumoai.graph.Edge`,
                or an :class:`~kumoai.graph.Edge` object itself.
            **kwargs: Any keyword arguments to construct an
                :class:`~kumoai.graph.Edge`.

        Raises:
            ValueError: if the edge is already present in the graph, if the
                source table does not exist in the graph, if the destination
                table does not exist in the graph, if the source key does not
                exist in the source table, or if the primary key of the source
                table is being treated as a foreign key.
        """
        edge = Edge._cast(*args, **kwargs)
        if edge is None:
            raise ValueError("Cannot add a 'None' edge to a graph.")

        (src_table, fkey, dst_table) = edge

        if edge in self.edges:
            raise ValueError(f"Cannot add edge {edge} to graph; edge is "
                             f"already present.")

        if not self.has_table(src_table):
            raise ValueError(
                f"Source table '{src_table}' does not exist in the graph.")

        if not self.has_table(dst_table):
            raise ValueError(
                f"Destination table '{dst_table}' does not exist in the "
                f"graph.")

        if not self.table(src_table).has_column(fkey):
            raise ValueError(
                f"Source key '{fkey}' does not exist in source table "
                f"'{src_table}'; please check that you have added it as a "
                f"column.")

        # Backend limitations: ensure the source is not its primary key:
        src_pkey = self.table(src_table).primary_key
        src_is_pkey = src_pkey is not None and src_pkey.name == fkey
        if src_is_pkey:
            raise ValueError(f"Cannot treat the primary key of table "
                             f"'{src_table}' as a foreign key; please "
                             f"select a different key.")

        self._edges.append(edge)
        return self

    def unlink(self, *args: Optional[Union[str, Edge]],
               **kwargs: str) -> 'LocalGraph':
        r"""Removes an :class:`~kumoai.graph.Edge` from the graph.

        Args:
            *args: Any arguments to construct an :class:`~kumoai.graph.Edge`,
                or an :class:`~kumoai.graph.Edge` object itself.
            **kwargs: Any keyword arguments to construct an
                :class:`~kumoai.graph.Edge`.

        Raises:
            ValueError: if the edge is not present in the graph.
        """
        edge = Edge._cast(*args, **kwargs)
        if edge not in self.edges:
            raise ValueError(f"Edge {edge} is not present in {self.edges}")
        self._edges.remove(edge)
        return self

    def infer_links(self) -> 'LocalGraph':
        r"""Infers links for the tables and adds them as edges to the graph.

        Note:
            This function expects graph edges to be undefined upfront.

        Raises:
            ValueError: If edges are not empty.
        """
        if self.edges is not None and len(self.edges) > 0:
            raise ValueError(
                "Cannot infer links if graph edges are not empty.")

        # TODO(blaz): expand logic for more elaborate inference
        self._infer_links_by_name()
        return self

    def _infer_links_by_name(self) -> 'LocalGraph':
        r"""Infers links for the tables in the graph by matching column
        names.
        """
        pk_index: DefaultDict[str, List[str]] = defaultdict(list)
        for table_name, table in self.tables.items():
            if table.has_primary_key():
                assert table.primary_key is not None
                pk_index[table.primary_key.name].append(table_name)

        # go through all non-pk columns and look up in the index
        for src_name, src in self.tables.items():
            src_pk_name: Optional[str] = (pk.name if (pk := src.primary_key)
                                          is not None else None)

            for col in src.columns:
                col_name = col.name

                # Skip the source PK itself
                if col_name == src_pk_name:
                    continue

                for dst_name in pk_index.get(col_name, ()):
                    self.link(Edge(src_name, col_name, dst_name))

        return self

    # Metadata ################################################################

    def validate(self) -> 'LocalGraph':
        r"""Validates the graph to ensure that all relevant metadata is
        specified for its tables and edges.

        Concretely, validation ensures that all tables are valid (see
        :meth:`kumoai.experimental.rfm.LocalTable.validate` for more
        information), and that edges properly link primary keys and foreign
        keys between valid tables.
        It additionally ensures that primary and foreign keys between tables
        in an :class:`~kumoai.graph.Edge` are of the same data type.

        Example:
            >>> import kumoai
            >>> graph = kumoai.LocalGraph(...)  # doctest: +SKIP
            >>> graph.validate()  # doctest: +SKIP
            ValueError: ...

        Raises:
            ValueError: if validation fails.
        """
        # TODO(blaz): lift restriction on isolated tables & conected components
        # validate tables
        for table in self.tables.values():
            table.validate()

        # validate edges
        self._validate_edges()

        # validate isolated tables and isolated components
        if len(self.tables) > 1:  # Only check if we have multiple tables
            self._validate_isolated_tables()
            self._validate_single_connected_component()

        return self

    def _validate_edges(self) -> None:
        r"""Validates that all edges in the graph are valid.

        Raises:
            ValueError: if the graph has invalid edges.
        """
        # validate edges fkey -> pkey
        for edge in self.edges:
            (src_table, fkey, dst_table) = edge
            dst_pkey = self.table(dst_table).primary_key
            # check that fkey exists in src_table

            if not self.table(src_table).has_column(fkey):
                raise ValueError(f"Edge {edge} is invalid as source table "
                                 f"'{src_table}' does not have a column with "
                                 f"name '{fkey}'.")
            # check that fkey is the primary key of dst_table
            if dst_pkey is None:
                raise ValueError(f"Edge {edge} is invalid since table "
                                 f"'{dst_table}' does not have a primary key.")

            # check that fkey/pkey are from either numerical or string family
            if (dtype_to_family(self.table(src_table)[fkey].dtype) not in [
                    "number", "string"
            ] or dtype_to_family(dst_pkey.dtype) not in ["number", "string"]):
                raise ValueError(f"Edge {edge} is invalid as foreign key "
                                 f"'{fkey}' and primary key '{dst_pkey.name}' "
                                 f"must be from be either 'numbers' or "
                                 f"'strings' (fkey.dtype: "
                                 f"{self.table(src_table)[fkey].dtype}, "
                                 f"pkey.dtype: {dst_pkey.dtype})")

            # check that fkey and pkey are from the same data type family
            if (dtype_to_family(self.table(src_table)[fkey].dtype)
                    != dtype_to_family(dst_pkey.dtype)):
                raise ValueError(f"Edge {edge} is invalid as foreign key "
                                 f"'{fkey}' and primary key '{dst_pkey.name}' "
                                 f"have incompatible data types. "
                                 f"(fkey.dtype: "
                                 f"{self.table(src_table)[fkey].dtype} "
                                 f"!= pkey.dtype: {dst_pkey.dtype})")

    def _validate_isolated_tables(self) -> None:
        r"""Validates that all tables in the graph are connected to at least
        one other table.

        Raises:
            ValueError: if the graph has isolated tables.
        """
        connected_tables = {edge.src_table for edge in self.edges} | \
            {edge.dst_table for edge in self.edges}

        for table_name in self.tables:
            if table_name not in connected_tables:
                raise ValueError(f"Table '{table_name}' is not connected to "
                                 f"any other tables in the graph (no incoming "
                                 f"or outgoing edges).")

    def _validate_single_connected_component(self) -> None:
        r"""Validates that all tables in the graph form a single connected
        component using breadth-first search.

        Raises:
            ValueError: if the graph has multiple disconnected components.
        """
        if not self.tables:
            return

        # Build adjacency list for undirected graph
        adjacency: Dict[str, Set[str]] = {
            table_name: set()
            for table_name in self.tables.keys()
        }
        for edge in self.edges:
            src, _, dst = edge
            adjacency[src].add(dst)
            adjacency[dst].add(src)  # Treat edges as bidirectional

        # BFS to find all reachable tables from the first table
        start_table = next(iter(self.tables.keys()))
        visited = set()
        queue = [start_table]
        visited.add(start_table)

        while queue:
            current = queue.pop(0)
            for neighbor in adjacency[current]:
                if neighbor not in visited:
                    visited.add(neighbor)
                    queue.append(neighbor)

        # Check if all tables are reachable
        unreachable_tables = set(self.tables.keys()) - visited
        if unreachable_tables:
            raise ValueError(
                f"Graph contains disconnected components. Tables "
                f"{list(unreachable_tables)} are not reachable from the main "
                f"component. All tables must be connected in a single "
                f"component.")

    # Class properties ########################################################

    def __hash__(self) -> int:
        return hash((tuple(self.edges), tuple(sorted(self.tables.keys()))))

    def __contains__(self, name: str) -> bool:
        return self.has_table(name)

    def __getitem__(self, name: str) -> LocalTable:
        return self.table(name)

    def __repr__(self) -> str:
        table_names = str(list(self.tables.keys())).replace("'", "")
        return (f'{self.__class__.__name__}(\n'
                f'  tables={table_names},\n'
                f'  edges={self.edges},\n'
                f')')

    # Conversion helpers ######################################################

    def _calculate_total_table_size_mb(self) -> float:
        """Calculate the total size of all tables in MB by
        temporarily serializing to parquet.

        Returns:
            Total size of all tables in megabytes
        """
        total_size_bytes = 0
        with tempfile.TemporaryDirectory() as tmp_dir:
            for table_name, local_table in self.tables.items():
                tmp_path = os.path.join(tmp_dir,
                                        f'{table_name}_size_check.parquet')
                local_table._data.to_parquet(tmp_path, index=False)
                total_size_bytes += os.path.getsize(tmp_path)

        return total_size_bytes / (1024 * 1024)  # Convert to MB

    def _upload_tables(self) -> None:
        """Upload LocalTable data using FileUploadConnector"""
        # TODO(blaz): upload with table_name_fpid instead of table_name
        with tempfile.TemporaryDirectory() as tmp_dir:
            for table_name, local_table in self.tables.items():
                tmp_path = os.path.join(tmp_dir, f'{table_name}.parquet')
                local_table._data.to_parquet(tmp_path, index=False)
                try:
                    upload_table(name=table_name, path=tmp_path)
                except Exception as e:
                    raise e

    def to_kumo_graph(self) -> Graph:
        """Upload tables and convert LocalGraph to kumo.graph.Graph

        This method handles both uploading the table data and converting
        the LocalGraph to a kumo Graph object.

        Returns:
            A kumo Graph object ready for use

        Raises:
            ValueError: If the total size of all tables exceeds 10GB
        """
        # Check total size before uploading
        total_size_mb = self._calculate_total_table_size_mb()

        if total_size_mb > MAX_TABLE_SIZE_MB:
            raise ValueError(
                f"Total size of all tables ({total_size_mb:.2f} MB) exceeds "
                f"the maximum allowed size of {MAX_TABLE_SIZE_MB} MB (10 GB). "
                f"Please reduce the data size before uploading.")

        self._upload_tables()
        return self._convert_to_kumo_graph()

    def _convert_to_kumo_graph(self) -> Graph:
        """Convert LocalGraph to kumo.graph.Graph using uploaded tables"""
        connector = FileUploadConnector(file_type="parquet")
        kumo_tables: Dict[str, Table] = {}

        for table_name, local_table in self.tables.items():
            source_table = connector[table_name]

            primary_key = (local_table.primary_key.name
                           if local_table.primary_key is not None else None)
            time_column = (local_table.time_column.name
                           if local_table.time_column is not None else None)
            kumo_table = Table.from_source_table(
                source_table=source_table,
                primary_key=primary_key,
                time_column=time_column,
            )

            # apply column metadata from local table to kumo table
            for local_col in local_table.columns:
                if kumo_table.has_column(local_col.name):
                    kumo_col = kumo_table.column(local_col.name)
                    if local_col.dtype:
                        kumo_col.dtype = local_col.dtype
                    if local_col.stype:
                        kumo_col.stype = local_col.stype

            kumo_tables[table_name] = kumo_table

        kumo_graph = Graph(tables=kumo_tables, edges=list(self.edges))

        return kumo_graph



import copy
import logging
import time
from typing import Any, Dict, List, Optional, Union

import kumoapi.data_snapshot as snapshot_api
import kumoapi.table as api
import pandas as pd
from kumoapi.common import JobStatus
from kumoapi.data_snapshot import TableSnapshotID
from kumoapi.typing import Stype
from typing_extensions import Self

from kumoai import global_state
from kumoai.client.table import TableID
from kumoai.connector import SourceColumn, SourceTable
from kumoai.graph.column import Column

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL_S = 20


class Table:
    r"""A Table represents metadata information for a table in a Kumo
    :class:`~kumoai.graph.Graph`.

    Whereas a :class:`~kumoai.connector.SourceTable` is simply a reference to a
    table behind a backing :class:`~kumoai.connector.Connector`, a table fully
    specifies the relevant metadata (including selected source columns, column
    data type and semantic type, and relational constraint information)
    necessary to train a :class:`~kumoai.pquery.PredictiveQuery` on graph of
    tables. A table can either be constructed explicitly, or with the
    convenience method :meth:`~kumoai.graph.Table.from_source_table`.

    .. code-block:: python

        import kumoai

        # Define connector to source data:
        connector = kumoai.S3Connector('s3://...')

        # Create table using `from_source_table`:
        customer = kumoai.Table.from_source_table(
            source_table=connector['customer'],
            primary_key='CustomerID',
        )

        # Create a table by constructing it directly:
        customer = kumoai.Table(
            source_table=connector['customer'],
            columns=[kumoai.Column(name='CustomerID', dtype='string', stype='ID')],
            primary_key='CustomerID',
        )

        # Infer any missing metadata in the table, from source table
        # properties:
        print("Current metadata: ", customer.metadata)
        customer.infer_metadata()

        # Validate the table configuration, for use in Kumo downstream models:
        customer.validate(verbose=True)

        # Fetch statistics from a snapshot of this table (this method will
        # take a table snapshot, and as a result may have high latency):
        customer.get_stats(wait_for="minimal")

    Args:
        source_table: The source table this Kumo table is created from.
        columns: The selected columns of the source table that are part of this
            Kumo table. Note that each column must specify its data type and
            semantic type; see the :class:`~kumoai.graph.Column` documentation
            for more information. If `None`  all columns from the
            source table are included by default.
        primary_key: The primary key of the table, if present. The primary key
            must exist in the :obj:`columns` argument.
        time_column: The time column of the table, if present. The time column
            must exist in the :obj:`columns` argument.
        end_time_column: The end time column of the table, if present. The end
            time column must exist in the :obj:`columns` argument.
    """  # noqa: E501

    def __init__(
        self,
        source_table: SourceTable,
        columns: Optional[List[Union[SourceColumn, Column]]] = None,
        primary_key: Optional[str] = None,
        time_column: Optional[str] = None,
        end_time_column: Optional[str] = None,
    ) -> None:
        # Reference to the source (raw) table:
        self.source_table = source_table
        self.source_name = source_table.name

        # Columns. Note that there is no distinction between columns treated as
        # features and those treated as constraints at this stage. The
        # treatment of columns as "feature" or "schema-only" columns will be
        # decided at the model plan stage (e.g. by encoding as `Null()`):
        self._columns: Dict[str, Column] = {}

        # Basic schema. This information is defined at the table level:
        self._primary_key: Optional[str] = None
        self._time_column: Optional[str] = None
        self._end_time_column: Optional[str] = None

        # Update values:
        if columns is None:
            columns = list(source_table.column_dict.values())
        for col in (columns or []):
            if isinstance(col, SourceColumn):
                col = Column(name=col.name, stype=col.stype, dtype=col.dtype)
            self.add_column(Column._cast(col))
        self.primary_key = Column._cast(primary_key)
        self.time_column = Column._cast(time_column)
        self.end_time_column = Column._cast(end_time_column)

        # Cached from backend. Note there is no such thing as a table resource
        # as tables are only persisted in the context of a graph. However,
        # table snapshot resources exist, as tables can be ingested and have
        # data fetched:
        self._table_snapshot_id: Optional[TableSnapshotID] = None

    @staticmethod
    def from_source_table(
        source_table: SourceTable,
        column_names: Optional[List[str]] = None,
        primary_key: Optional[str] = None,
        time_column: Optional[str] = None,
        end_time_column: Optional[str] = None,
    ) -> 'Table':
        r"""Creates a Kumo Table from a source table. If no column names are
        specified, all source columns are included by default.

        Args:
            source_table: The :class:`~kumoai.connector.SourceTable` object
                that this table is constructed on.
            column_names: A list of columns to include from the source table;
                if not specified, all columns are included by default.
            primary_key: The name of the primary key of this table, if it
                exists.
            time_column: The name of the time column of this table, if it
                exists.
            end_time_column: The name of the end time column of this table, if
                it exists.
        """
        cols = [
            Column(name, col.stype, col.dtype)
            for name, col in source_table.column_dict.items()
            if (name in column_names if column_names is not None else True)
        ]
        out = Table(source_table, cols)
        out.primary_key = Column._cast(primary_key)
        out.time_column = Column._cast(time_column)
        out.end_time_column = Column._cast(end_time_column)
        return out

    def print_definition(self) -> None:
        r"""Prints the full definition for this table; this definition can be
        copied-and-pasted verbatim to re-create this table.
        """
        pkey_name = (f"\"{self.primary_key.name}\""
                     if self.primary_key is not None else "None")
        t_name = (f"\"{self.time_column.name}\""
                  if self.time_column is not None else "None")
        et_name = (f"\"{self.end_time_column.name}\""
                   if self.end_time_column is not None else "None")
        col_dict = "\n".join([f'    {c},' for c in self.columns])
        source_repr = f"{self.source_table.connector}[\"{self.source_name}\"]"
        print(f'{self.__class__.__name__}(\n'
              f'  source_table={source_repr},\n'
              f'  primary_key={pkey_name},\n'
              f'  time_column={t_name},\n'
              f'  end_time_column={et_name},\n'
              f'  columns=[\n{col_dict}\n'
              f'  ],\n'
              f')')

    # Data column #############################################################

    def has_column(self, name: str) -> bool:
        r"""Returns True if this table has column with name :obj:`name`; False
        otherwise.
        """
        return name in self._columns

    def column(self, name: str) -> Column:
        r"""Returns the data column named with name :obj:`name` in this table,
        or raises a :obj:`KeyError` if no such column is present.

        Raises:
            :class:`KeyError`
                if :obj:`name` is not present in this table.
        """
        if not self.has_column(name):
            raise KeyError(
                f"Column '{name}' not found in table '{self.source_name}'")
        return self._columns[name]

    @property
    def columns(self) -> List[Column]:
        r"""Returns a list of :class:`~kumoai.Column` objects that represent
        the columns in this table.
        """
        return list(self._columns.values())

    def add_column(self, *args: Any, **kwargs: Any) -> None:
        r"""Adds a :obj:`~kumoai.graph.Column` to this table. A column can
        either be added by directly specifying its configuration in this call,
        or by creating a Column object and passing it as an argument.

        Example:
            >>> import kumoai
            >>> table = kumoai.Table(source_table=...)  # doctest: +SKIP
            >>> table.add_column(name='col1', dtype='string')  # doctest: +SKIP
            >>> table.add_column(kumoai.Column('col2', 'int'))  # doctest: +SKIP

        .. # noqa: E501
        """
        col = Column._cast(*args, **kwargs)
        if col is None:
            raise ValueError("Cannot add a 'None' column to a table.")
        if self.has_column(col.name):
            self._columns[col.name].update(col)
        else:
            self._columns[col.name] = col

    def remove_column(self, name: str) -> Self:
        r"""Removes a :obj:`~kumoai.graph.Column` from this table.

        Raises:
            :class:`KeyError`
                if :obj:`name` is not present in this table.
        """
        if not self.has_column(name):
            raise KeyError(
                f"Column '{name}' not found in table '{self.source_name}'")

        if self.has_primary_key() and self._primary_key == name:
            self.primary_key = None
        if self.has_time_column() and self._time_column == name:
            self.time_column = None
        if self.has_end_time_column() and self._end_time_column == name:
            self.end_time_column = None
        del self._columns[name]
        return self

    # Primary key #############################################################

    def has_primary_key(self) -> bool:
        r"""Returns :obj:`True` if this table has a primary key; :obj:`False`
        otherwise.
        """
        return self._primary_key is not None

    @property
    def primary_key(self) -> Optional[Column]:
        r"""The primary key column of this table.

        The getter returns the primary key column of this table, or None if no
        such primary key is present.

        The setter sets a column as a primary key on this table, and raises a
        :class:`ValueError` if the primary key has a non-ID semantic type.
        """
        if not self.has_primary_key():
            return None
        assert self._primary_key is not None
        return self._columns[self._primary_key]

    @primary_key.setter
    def primary_key(self, *args: Any, **kwargs: Any) -> Self:
        col = Column._cast(*args, **kwargs)
        if col is None:
            self._primary_key = None
            return self

        if col.stype is not None and col.stype != Stype.ID:
            raise ValueError(
                f"The semantic type of a primary key must be 'ID' (got "
                f"{col.stype}).")

        col.stype = Stype.ID
        self.add_column(col)
        self._primary_key = col.name
        return self

    # Time column #############################################################

    def has_time_column(self) -> bool:
        r"""Returns :obj:`True` if this table has a time column; :obj:`False`
        otherwise.
        """
        return self._time_column is not None

    @property
    def time_column(self) -> Optional[Column]:
        r"""The time column of this table.

        The getter returns the time column of this table, or :obj:`None` if no
        such time column is present.

        The setter sets a column as a time column on this table, and raises a
        :class:`ValueError` if the time column is the same as the end time
        column, or has a non-timestamp semantic type.
        """
        if not self.has_time_column():
            return None
        assert self._time_column is not None
        return self._columns[self._time_column]

    @time_column.setter
    def time_column(self, *args: Any, **kwargs: Any) -> Self:
        col = Column._cast(*args, **kwargs)
        if col is None:
            self._time_column = None
            return self

        if self.has_end_time_column() and self._end_time_column == col.name:
            raise ValueError(f"Cannot set the time column ('{col.name}') "
                             f"to be the same as the end time column "
                             f"('{self._end_time_column}')")

        if col.stype is not None and col.stype != Stype.timestamp:
            raise ValueError(
                f"The semantic type of a time column must be 'timestamp' (got "
                f"{col.stype}).")

        col.stype = Stype.timestamp
        self.add_column(col)
        self._time_column = col.name
        return self

    # End time column #########################################################

    def has_end_time_column(self) -> bool:
        r"""Returns :obj:`True` if this table has an end time column;
        :obj:`False` otherwise.
        """
        return self._end_time_column is not None

    @property
    def end_time_column(self) -> Optional[Column]:
        r"""The end time column of this table.

        The getter returns the end time column of this table, or :obj:`None` if
        no such column is present.

        The setter sets a column as a time column on this table, and raises a
        :class:`ValueError` if the time column is the same as the end time
        column, or has a non-timestamp semantic type.
        """
        if not self.has_end_time_column():
            return None
        assert self._end_time_column is not None
        return self._columns[self._end_time_column]

    @end_time_column.setter
    def end_time_column(self, *args: Any, **kwargs: Any) -> Self:
        col = Column._cast(*args, **kwargs)
        if col is None:
            self._end_time_column = None
            return self

        if self.has_time_column() and self._time_column == col.name:
            raise ValueError(f"Cannot set the end time column ('{col.name}') "
                             f"to be the same as the time column "
                             f"('{self._time_column}')")

        if col.stype is not None and col.stype != Stype.timestamp:
            raise ValueError(
                f"The semantic type of an end time column must be 'timestamp' "
                f"(got {col.stype}).")

        col.stype = Stype.timestamp
        self.add_column(col)
        self._end_time_column = col.name
        return self

    # Metadata ################################################################

    @property
    def metadata(self) -> pd.DataFrame:
        r"""Returns a :class:`~pandas.DataFrame` object containing Kumo metadata
        information about the columns in this table.

        The returned dataframe has columns ``name``, ``dtype``, ``stype``,
        ``is_primary_key``, ``is_time_column``, and ``is_end_time_column``,
        which provide an aggregate view of the properties of the columns of
        this table.

        Example:
            >>> import kumoai
            >>> table = kumoai.Table(source_table=...)  # doctest: +SKIP
            >>> table.add_column(name='CustomerID', dtype='float64', stype='ID')  # doctest: +SKIP
            >>> table.metadata  # doctest: +SKIP
                name        dtype       stype    is_time_column  is_end_time_column
            0   CustomerID  float64     ID       False           False
        """  # noqa: E501
        items = self._columns.items()
        col_names: List[str] = [i[0] for i in items]
        cols: List[Column] = [i[1] for i in items]

        return pd.DataFrame({
            'name':
            pd.Series(dtype=str, data=col_names),
            'dtype':
            pd.Series(
                dtype=str, data=[
                    c.dtype.value if c.dtype is not None else None
                    for c in cols
                ]),
            'stype':
            pd.Series(
                dtype=str, data=[
                    c.stype.value if c.stype is not None else None
                    for c in cols
                ]),
            'is_primary_key':
            pd.Series(dtype=bool, data=[self.primary_key == c for c in cols]),
            'is_time_column':
            pd.Series(dtype=bool, data=[self.time_column == c for c in cols]),
            'is_end_time_column':
            pd.Series(dtype=bool,
                      data=[self.end_time_column == c for c in cols]),
        })

    def infer_metadata(self, inplace: bool = True) -> Self:
        r"""Infers all metadata for this table's specified columns, including
        the column data types, semantic types, timestamp formats, primary keys,
        and time/end-time columns

        Args:
            inplace: Whether the method should modify the table columns in
                place or return a new :class:`~kumoai.graph.Table` object.

        .. note::
            This method in-place modifies the Table object if `inplace = True`,
            and returns a copy if ``inplace = False``.
        """
        col_requests: List[api.ColumnMetadataRequest] = []
        for col in self.columns:
            col_requests.append(
                # stype and dtype are None to support inferral:
                api.ColumnMetadataRequest(
                    name=col.name,
                    stype=None,
                    dtype=None,
                    timestamp_format=col.timestamp_format,
                ))

        pk_name: Optional[str] = None
        if self.has_primary_key():
            pk_name = self.primary_key.name  # type: ignore

        tc_name: Optional[str] = None
        if self.has_time_column():
            tc_name = self.time_column.name  # type: ignore

        request = api.TableMetadataRequest(
            cols=col_requests,
            source_table=self.source_table._to_api_source_table(),
            pkey=pk_name,
            time_col=tc_name,
        )

        response = global_state.client.table_api.infer_metadata(request)
        inferred_cols: Dict[str, api.Column] = {
            col.name: col
            for col in response.cols
        }

        # Handle inplace:
        out = self
        if not inplace:
            out = copy.deepcopy(self)

        # TODO(manan): respect user overrides
        # TODO(manan): what happens when the ts format is set based on an
        # override?
        for col in out.columns:
            inferred_col = inferred_cols[col.name]

            col.dtype = inferred_col.dtype
            col.stype = inferred_col.stype
            col.timestamp_format = (col.timestamp_format
                                    or inferred_col.timestamp_format)

        # TODO(manan): support end-time column
        if not out.has_primary_key() and response.pkey is not None:
            out.primary_key = response.pkey
        if not out.has_time_column() and response.time_col is not None:
            out.time_column = response.time_col

        # Override for Kumo backend, always:
        if out.has_primary_key():
            out.primary_key.stype = Stype.ID  # type: ignore

        if out.has_time_column():
            out.time_column.stype = Stype.timestamp  # type: ignore

        if out.has_end_time_column():
            out.end_time_column.stype = Stype.timestamp  # type: ignore

        return out

    def _validate_definition(self) -> None:
        for col in self.columns:
            if col.dtype is None or col.stype is None:
                raise ValueError(
                    f"Column {col.name} is not fully specified. Please "
                    f"specify this column's data type and semantic type "
                    f"before proceeding. {col.name} currently has a "
                    f"data type of {col.dtype} and semantic type of "
                    f"{col.stype}.")

    def validate(self, verbose: bool = True) -> Self:
        r"""Validates a Table to ensure that all relevant metadata is specified
        for a table to be used in a downstream :class:`~kumoai.graph.Graph` and
        :class:`~kumoai.pquery.PredictiveQuery`.

        Conceretely, validation ensures that all columns have valid
        data and semantic types, with respect to the table's source data.
        For example, if a text column is assigned a ``dtype`` of ``"int"``,
        this method will raise an exception detailing the mismatch. Similarly,
        if a column cannot be cast from its source data type to the specified
        data type (*e.g* ``"int"`` to ``"binary"``), this method will raise an
        exception.

        .. warning::
            Data type validation is performed on a sample of table data. A
            valid response may not indicate your entire data source is
            configured correctly.

        Args:
            verbose: Whether to log non-error output of this validation.

        Example:
            >>> import kumoai
            >>> table = kumoai.Table(...)  # doctest: +SKIP
            >>> table.validate()  # doctest: +SKIP

        Raises:
            ValueError:
                if validation fails.
        """
        self._validate_definition()

        # Actual heavy lifting:
        resp = global_state.client.table_api.validate_table(
            api.TableValidationRequest(self._to_api_table_definition()))
        if not resp.ok:
            raise ValueError(resp.error_message())
        if verbose:
            if resp.empty():
                logger.info("Table %s is configured correctly.",
                            self.source_name)
            else:
                logger.warning(resp.message())
        return self

    # Snapshot ################################################################

    @property
    def snapshot_id(self) -> Optional[snapshot_api.TableSnapshotID]:
        r"""Returns the snapshot ID of this table's snapshot, if a snapshot
        has been taken. Returns `None` otherwise.

        .. warning::
            This property currently only returns a snapshot ID if a snapshot
            has been taken *in this session.*
        """
        return self._table_snapshot_id

    def snapshot(
        self,
        *,
        force_refresh: bool = False,
        non_blocking: bool = False,
    ) -> snapshot_api.TableSnapshotID:
        r"""Takes a *snapshot* of this table's underlying data, and returns a
        unique identifier for this snapshot.

        The *snapshot* functionality allows one to freeze a table in time, so
        that underlying data changes do not require Kumo to re-process the
        data. This allows for fast iterative machine learning model
        development, on a consistent set of input data.

        .. warning::
            Please note that snapshots are intended to freeze tables in
            time, and not to allow for "time-traveling" to an earlier version
            of data with a prior snapshot. In particular, this means that a
            table can only have one version of a snapshot, which represents
            the latest snapshot taken for that table.

        .. note::
            If you are using Kumo as a Snowpark Container Services native
            application, please note that *snapshot* is a no-op for all
            non-view tables.

        Args:
            force_refresh: Indicates whether a snapshot should be taken, if one
                already exists in Kumo. If :obj:`False`, a previously existing
                snapshot may be re-used. If :obj:`True`, a new snapshot is
                always taken.
            non_blocking: Whether this operation should return immediately
                after creating the snapshot, or await completion of the
                snapshot. If :obj:`True`, the snapshot will proceed in the
                background, and will be used for any downstream job.
        """
        if self._table_snapshot_id is None or force_refresh:
            self._table_snapshot_id = (
                global_state.client.table_api.create_snapshot(
                    table_definition=self._to_api_table_definition(),
                    refresh_source=True,
                ))

            stage = snapshot_api.TableSnapshotStage.INGEST
            resource: snapshot_api.TableSnapshotResource = (
                global_state.client.table_api.get_snapshot(
                    snapshot_id=self._table_snapshot_id))

            if not non_blocking:
                status = resource.stages[stage].status
                while not status.is_terminal:
                    # TODO(manan, siyang): fix start and end time
                    resource = (global_state.client.table_api.get_snapshot(
                        snapshot_id=self._table_snapshot_id))
                    logger.info(
                        "Awaiting snapshot completion: current status is %s ",
                        status)
                    time.sleep(_DEFAULT_INTERVAL_S)
                    status = resource.stages[stage].status

                state = resource.stages[stage]
                status = state.status
                warnings = "\n".join([
                    f"{i}. {message}"
                    for i, message in enumerate(state.warnings)
                ])
                error = state.error
                if status == JobStatus.FAILED:
                    raise RuntimeError(
                        f"Table snapshot with identifier "
                        f"{self._table_snapshot_id} failed, with error "
                        f"{error} and warnings {warnings}")
                if len(state.warnings) > 0:
                    logger.warning(
                        "Table snapshot completed with the following "
                        "warnings: %s", warnings)

        # <prefix>@<data_version>
        assert self._table_snapshot_id is not None
        return self._table_snapshot_id

    def get_stats(
        self,
        wait_for: Optional[str] = None,
    ) -> Dict[str, Dict[str, Any]]:
        r"""Returns all currently computed statistics on the latest snapshot of
        this table. If a snapshot on this table has not been taken, this method
        will take a snapshot.

        .. note::
            Table statstics are computed in multiple stages after ingestion is
            complete. These stages are called *minimal* and *full*; minimal
            statistics are always computed before full statistics.

        Args:
            wait_for: Whether this operation should block on the existence of
                statistics availability. This argument can take one of three
                values: :obj:`None`, which indicates that the method should
                return immediately with whatever statistics are present,
                :obj:`"minimal"`, which indicates that the method should return
                the when the minimum, maximum, and fraction of NA values
                statistics are present, or :obj:`"full"`, which indicates that
                the method should return when all computed statistics are
                present.
        """
        assert wait_for is None or wait_for in {"minimal", "full"}

        # Attempt to snapshot, use cached snapshot if possible:
        if not self._table_snapshot_id:
            self.snapshot(force_refresh=False, non_blocking=False)
        assert self._table_snapshot_id is not None

        # Fetch resource:
        resource: snapshot_api.TableSnapshotResource = (
            global_state.client.table_api.get_snapshot(
                snapshot_id=self._table_snapshot_id))

        # Wait for a stage, if we need to:
        if wait_for:
            if wait_for == "minimal":
                stage = snapshot_api.TableSnapshotStage.MIN_COL_STATS
            else:
                stage = snapshot_api.TableSnapshotStage.FULL_COL_STATS

            status = resource.stages[stage].status
            while not status.is_terminal:
                resource = (global_state.client.table_api.get_snapshot(
                    snapshot_id=self._table_snapshot_id))
                logger.info(
                    "Awaiting %s column statistics: current status is %s ",
                    wait_for, status)
                time.sleep(_DEFAULT_INTERVAL_S)
                status = resource.stages[stage].status

        # Write out statistics:
        out = {}
        col_stats = resource.column_stats
        for stat in (col_stats or []):
            out[stat.column_name] = stat.stats
        return out

    # Persistence #############################################################

    def _to_api_table_definition(self) -> api.TableDefinition:
        # TODO(manan): type narrowing?
        pk_name: Optional[str] = None
        if self.has_primary_key():
            pk_name = self.primary_key.name  # type: ignore

        tc_name: Optional[str] = None
        if self.has_time_column():
            tc_name = self.time_column.name  # type: ignore

        etc_name: Optional[str] = None
        if self.has_end_time_column():
            etc_name = self.end_time_column.name  # type: ignore

        return api.TableDefinition(
            cols=[
                api.Column(col.name, col.stype, col.dtype,
                           col.timestamp_format) for col in self.columns
            ],
            source_table=self.source_table._to_api_source_table(),
            pkey=pk_name,
            time_col=tc_name,
            end_time_col=etc_name,
        )

    @staticmethod
    def _from_api_table_definition(
            table_definition: api.TableDefinition) -> 'Table':
        return Table(
            source_table=SourceTable._from_api_table_definition(
                table_definition),
            columns=[
                Column(col.name, col.stype, col.dtype, col.timestamp_format)
                for col in table_definition.cols
            ],
            primary_key=table_definition.pkey,
            time_column=table_definition.time_col,
            end_time_column=table_definition.end_time_col,
        )

    def save(self, name: Optional[str] = None) -> Union[TableID, str]:
        r"""Associates this table with a unique name, that can later be
        used to fetch the table either in the Kumo UI or in the Kumo SDK
        with method :meth:`~kumoai.Table.load`.

        Args:
            name: The name to associate with this table definition. If the
                name is already associated with another table, that table will
                be overridden.

        Example:
            >>> import kumoai
            >>> table = kumoai.Table(...)  # doctest: +SKIP
            >>> unique_id = table.save()  # doctest: +SKIP
            >>> loaded = kumoai.Table.load(unique_id) # doctest: +SKIP
            >>> name = table.save("name")  # doctest: +SKIP
            >>> loaded = kumoai.Table.load("name")  # doctest: +SKIP
        """
        self.validate(verbose=False)
        template_resource = (global_state.client.table_api.get_table_if_exists(
            table_id_or_name=name)) if name else None

        if template_resource is not None:
            config = self._from_api_table_definition(template_resource.table)
            logger.warning(
                ("Table template %s already exists, with configuration %s. "
                 "This template will be overridden with configuration %s."),
                name, str(config), str(self))

        # TODO(manan): fix
        _id = global_state.client.table_api.create_table(
            table_def=self._to_api_table_definition(),
            name_alias=name,
            force_rename=True if name else False,
        )
        return f"table-{_id.split('-', maxsplit=1)[1]}"

    @classmethod
    def load(cls, table_id_or_template: str) -> 'Table':
        r"""Loads a table from either a table ID or a named template. Returns a
        :class:`Table` object that contains the loaded table along with its
        columns, etc.
        """
        api = global_state.client.table_api
        res = api.get_table_if_exists(table_id_or_template)
        if not res:
            raise ValueError(f"Table {table_id_or_template} was not found.")
        out = cls._from_api_table_definition(res.table)
        return out

    # Class properties ########################################################

    def __hash__(self) -> int:
        return hash(
            tuple(self.columns +
                  [self.primary_key, self.time_column, self.end_time_column]))

    def __contains__(self, name: str) -> bool:
        return self.has_column(name)

    def __getitem__(self, name: str) -> Column:
        return self.column(name)

    def __delitem__(self, name: str) -> None:
        self.remove_column(name)

    def __repr__(self) -> str:
        col_names = str(list(self._columns.keys())).replace("'", "")
        pkey_name = (self.primary_key.name
                     if self.primary_key is not None else "None")
        t_name = (self.time_column.name
                  if self.time_column is not None else "None")
        et_name = (self.end_time_column.name
                   if self.end_time_column is not None else "None")
        return (f'{self.__class__.__name__}(\n'
                f'  source_name={self.source_name},\n'
                f'  data_source={self.source_table.connector.name},\n'
                f'  columns={col_names},\n'
                f'  primary_key={pkey_name},\n'
                f'  time_column={t_name},\n'
                f'  end_time_column={et_name},\n'
                f')')



from dataclasses import dataclass, field
from enum import Enum
from typing import Final, Optional


class HTTPMethod(Enum):
    r"""HTTP methods supported by the API."""
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PATCH = "PATCH"


@dataclass(frozen=True)
class Endpoint:
    r"""Represents an API endpoint with its path and HTTP method."""
    path: Optional[str] = field(default=None)
    method: HTTPMethod = HTTPMethod.GET

    def validate(self) -> None:
        pass

    def get_path(self) -> str:
        if self.path is None:
            # This should be in validate but is here for type checker
            raise ValueError("Endpoint requires a path")
        return self.path


@dataclass(frozen=True)
class IDEndpoint(Endpoint):
    r"""Represents an API endpoint with an additional id/name in its path."""
    template_path: Optional[str] = field(default=None)

    def __post_init__(self) -> None:
        if self.template_path is None:
            raise ValueError("template_path must be set explicitly")
        if "{id}" not in self.template_path:
            raise ValueError("IDEndpoint path must contain '{{id}}': "
                             f"got '{self.template_path}'")

    def with_id(self, resource_id: str) -> 'IDEndpoint':
        assert self.template_path is not None
        return IDEndpoint(template_path=self.template_path,
                          path=self.template_path.format(id=resource_id),
                          method=self.method)

    def validate(self) -> None:
        if self.path is None:
            raise ValueError(
                "IDEndpoint requires with_id() to be called with a resource id"
            )

    @classmethod
    def from_base(cls, base: str, method: HTTPMethod) -> 'IDEndpoint':
        r"""Alternate constructor to build a template_path by appending /{id}
        to base.
        """
        template = f"{base}/{{id}}"
        return cls(template_path=template, method=method)


class ConnectorEndpoints:
    BASE: Final[str] = "/connectors"

    create = Endpoint(BASE, HTTPMethod.POST)
    list = Endpoint(BASE, HTTPMethod.GET)
    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)
    delete = IDEndpoint.from_base(BASE, HTTPMethod.DELETE)

    start_file_upload = Endpoint(f"{BASE}/utils/start_file_upload",
                                 HTTPMethod.POST)
    delete_uploaded_file = Endpoint(f"{BASE}/utils/delete_uploaded_file",
                                    HTTPMethod.POST)
    complete_file_upload = Endpoint(f"{BASE}/utils/complete_file_upload",
                                    HTTPMethod.POST)


class PQueryEndpoints:
    BASE: Final[str] = '/predictive_queries'

    create = Endpoint(BASE, HTTPMethod.POST)
    list = Endpoint(BASE, HTTPMethod.GET)
    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)
    delete = IDEndpoint.from_base(BASE, HTTPMethod.DELETE)

    infer_task_type = Endpoint(f"{BASE}/infer_task_type", HTTPMethod.POST)
    validate = Endpoint(f"{BASE}/validate", HTTPMethod.POST)
    suggest_training_table_plan = Endpoint(
        f"{BASE}/train_table_generation_plan", HTTPMethod.POST)
    suggest_model_plan = Endpoint(f"{BASE}/model_plan", HTTPMethod.POST)


class SourceTableEndpoints:
    BASE: Final[str] = "/source_tables"

    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)

    list_tables = Endpoint(f"{BASE}/list_tables", HTTPMethod.POST)
    validate_table = Endpoint(f"{BASE}/validate_table", HTTPMethod.POST)
    get_table_data = Endpoint(f"{BASE}/get_table_data", HTTPMethod.POST)
    get_table_config = Endpoint(f"{BASE}/get_table_config", HTTPMethod.POST)


class GraphEndpoints:
    BASE: Final[str] = "/graphs"
    SNAPSHOTS: Final[str] = "/graphsnapshots"

    create = Endpoint(BASE, HTTPMethod.POST)
    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)

    validate = Endpoint(f"{BASE}/validate", HTTPMethod.POST)
    infer_links = Endpoint(f"{BASE}/infer_links", HTTPMethod.POST)
    create_snapshot = Endpoint(SNAPSHOTS, HTTPMethod.POST)
    get_snapshot = IDEndpoint.from_base(SNAPSHOTS, HTTPMethod.GET)
    get_edge_stats = IDEndpoint.from_base(f"{SNAPSHOTS}/edge_health",
                                          method=HTTPMethod.GET)


class OnlineServingEndpoints:
    BASE: Final[str] = "/online_serving_endpoints"

    create = Endpoint(BASE, HTTPMethod.POST)
    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)
    list = Endpoint(BASE, HTTPMethod.GET)
    update = IDEndpoint.from_base(BASE, HTTPMethod.PATCH)
    delete = IDEndpoint.from_base(BASE, HTTPMethod.DELETE)


class TableEndpoints:
    BASE: Final[str] = "/tables"
    SNAPSHOTS: Final[str] = "/tablesnapshots"

    create = Endpoint(BASE, HTTPMethod.POST)
    get = IDEndpoint.from_base(BASE, HTTPMethod.GET)

    create_snapshot = Endpoint(SNAPSHOTS, HTTPMethod.POST)
    get_snapshot = IDEndpoint.from_base(SNAPSHOTS, HTTPMethod.GET)
    validate = Endpoint(f"{BASE}/validate", HTTPMethod.POST)
    infer_metadata = Endpoint(f"{BASE}/infer_metadata", HTTPMethod.POST)



from typing import List

from kumoapi.json_serde import to_json_dict
from kumoapi.source_table import (
    SourceTableConfigRequest,
    SourceTableConfigResponse,
    SourceTableDataRequest,
    SourceTableDataResponse,
    SourceTableListRequest,
    SourceTableListResponse,
    SourceTableValidateRequest,
    SourceTableValidateResponse,
)

from kumoai.client import KumoClient
from kumoai.client.endpoints import SourceTableEndpoints
from kumoai.client.utils import parse_response, raise_on_error


class SourceTableAPI:
    r"""Typed API definition for Kumo source tables."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client

    def validate_table(
            self, request: SourceTableValidateRequest
    ) -> SourceTableValidateResponse:
        response = self._client._request(SourceTableEndpoints.validate_table,
                                         json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(SourceTableValidateResponse, response)

    def list_tables(
            self, request: SourceTableListRequest) -> SourceTableListResponse:
        response = self._client._request(SourceTableEndpoints.list_tables,
                                         json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(SourceTableListResponse, response)

    def get_table_data(
            self, request: SourceTableDataRequest) -> SourceTableDataResponse:
        response = self._client._request(SourceTableEndpoints.get_table_data,
                                         json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(List[SourceTableDataResponse], response)

    def get_table_config(
            self,
            request: SourceTableConfigRequest) -> SourceTableConfigResponse:
        response = self._client._request(SourceTableEndpoints.get_table_config,
                                         json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(SourceTableConfigResponse, response)



import datetime
import warnings

import kumoai

warnings.filterwarnings(
    'ignore',
    message="The default value for 'linkcheck_report_timeouts_as_broken'",
    category=DeprecationWarning, module='sphinx')

author = 'Kumo.AI'
project = 'kumoai'
version = kumoai.__version__
copyright = f'{datetime.datetime.now().year}, {author}'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_copybutton',
    'sphinx_autodoc_typehints',
]

html_theme = 'sphinx_book_theme'
html_logo = '../assets/kumo-logo.svg'
html_favicon = '../assets/kumo-favicon.png'
html_theme_options = {
    'navigation_depth': 2,
    'extra_footer': f'<div>Version: {kumoai.__version__}</div>'
}
templates_path = ['_templates']
autodoc_member_order = 'bysource'
autodoc_class_signature = 'separated'
copybutton_prompt_text = r'>>> |\.\.\. '
copybutton_prompt_is_regexp = True

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pandas': ('http://pandas.pydata.org/pandas-docs/dev', None),
    'requests': ('https://requests.readthedocs.io/en/stable/', None),
}

doctest_global_setup = """
import os

os.environ['SNOWFLAKE_USER'] = '...'
os.environ['SNOWFLAKE_PASSWORD'] = '...'
"""


def rst_jinja_render(app, _, source):
    if hasattr(app.builder, 'templates'):
        rst_context = {'kumoai': kumoai}
        source[0] = app.builder.templates.render_string(source[0], rst_context)


def setup(app):
    app.connect('source-read', rst_jinja_render)



from http import HTTPStatus
from typing import Any, List, Optional

from kumoapi.json_serde import to_json_dict
from kumoapi.online_serving import (
    OnlineServingEndpointRequest,
    OnlineServingEndpointResource,
)

from kumoai.client import KumoClient
from kumoai.client.endpoints import OnlineServingEndpoints
from kumoai.client.utils import (
    parse_id_response,
    parse_patch_response,
    parse_response,
    raise_on_error,
)

OnlineServingEndpointID = str


class OnlineServingEndpointAPI:
    r"""Typed API definition for Kumo graph definition."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client
        self._base_endpoint = '/online_serving_endpoints'

    # TODO(blaz): document final interface
    def create(
        self,
        req: OnlineServingEndpointRequest,
        **query_params: Any,
    ) -> OnlineServingEndpointID:
        """Creates a new online serving endpoint.

        Args:
            req (OnlineServingEndpointRequest): request body.
            use_ge (Optional[bool], optional): If present, override graph
            backend option to use GRAPHENGINE if true else MEMORY.
            **query_params: Additional query parameters to pass to the API.

        Returns:
            OnlineServingEndpointID: unique endpoint resource id.
        """
        resp = self._client._post(
            self._base_endpoint,
            params=query_params if query_params else None,
            json=to_json_dict(req),
        )
        raise_on_error(resp)
        return parse_id_response(resp)

    def get_if_exists(
        self, id: OnlineServingEndpointID
    ) -> Optional[OnlineServingEndpointResource]:
        resp = self._client._request(OnlineServingEndpoints.get.with_id(id))
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise_on_error(resp)
        return parse_response(OnlineServingEndpointResource, resp)

    def list(self) -> List[OnlineServingEndpointResource]:
        resp = self._client._request(OnlineServingEndpoints.list)
        raise_on_error(resp)
        return parse_response(List[OnlineServingEndpointResource], resp)

    def update(self, id: OnlineServingEndpointID,
               req: OnlineServingEndpointRequest) -> bool:
        resp = self._client._request(OnlineServingEndpoints.update.with_id(id),
                                     data=to_json_dict(req))
        raise_on_error(resp)
        return parse_patch_response(resp)

    def delete(self, id: OnlineServingEndpointID) -> None:
        """This is idempotent and can be called multiple times."""
        resp = self._client._request(OnlineServingEndpoints.delete.with_id(id))
        raise_on_error(resp)



import re

import pandas as pd
from kumoapi.typing import Dtype, Stype

# Column names suffixes that end in "id" but should not be given the ID stype.
_IGNORED_ID_SUFFIXES = [
    'bid',
    'acid',
    'grid',
    'maid',
    'paid',
    'raid',
    'void',
    'avoid',
    'braid',
    'covid',
    'fluid',
    'rabid',
    'solid',
    'hybrid',
    'inlaid',
    'liquid',
]


def contains_id(ser: pd.Series, column_name: str, dtype: Dtype) -> bool:
    if not Stype.ID.supports_dtype(dtype):
        return False

    column_name = column_name.lower()

    match = re.search(
        '(^|_)(id|hash|key|code|uuid)(_|$)',
        column_name,
        re.IGNORECASE,
    )
    if match is not None:
        return True

    if not column_name.endswith('id'):
        return False
    for suffix in _IGNORED_ID_SUFFIXES:
        if column_name.endswith(suffix):
            return False
    return True



import functools
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from kumoapi.jobs import MetadataField, WriteMode

# Pydantic compatibility imports
try:
    from pydantic import __version__ as pydantic_version
    PYDANTIC_V2 = int(pydantic_version.split('.')[0]) >= 2
except ImportError:
    PYDANTIC_V2 = False

if PYDANTIC_V2:
    from pydantic import field_validator, model_validator
    from pydantic.dataclasses import dataclass as pydantic_dataclass
else:
    from pydantic import root_validator, validator
    from pydantic.dataclasses import dataclass as pydantic_dataclass

from kumoai.connector.base import Connector


def compatible_field_validator(field_name: str):  # type: ignore
    """Decorator factory that creates a field validator compatible with both
    Pydantic v1 and v2.

    Usage:
        @compatible_field_validator('field_name')
        def validate_field(cls, v, values_or_info):
            # Your validation logic here
            return v
    """
    def decorator(func):  # type: ignore
        if PYDANTIC_V2:

            @field_validator(field_name)
            @classmethod
            @functools.wraps(func)
            def wrapper(cls, v, info):  # type: ignore
                # Convert info to values dict for compatibility
                values = info.data if hasattr(info, 'data') else {}
                return func(cls, v, values)

            return wrapper
        else:

            @validator(field_name)
            @functools.wraps(func)
            def wrapper(cls, v, values):  # type: ignore
                return func(cls, v, values)

            return wrapper

    return decorator


def compatible_model_validator(mode='before'):  # type: ignore
    """Decorator factory that creates a model validator compatible with both
    Pydantic v1 and v2.

    Usage:
        @compatible_model_validator()
        def validate_model(cls, values):
            # Your validation logic here
            return values
    """
    def decorator(func):  # type: ignore
        if PYDANTIC_V2:

            @model_validator(mode=mode)
            @classmethod
            @functools.wraps(func)
            def wrapper(cls, values):  # type: ignore
                return func(cls, values)

            return wrapper
        else:

            @root_validator
            @functools.wraps(func)
            def wrapper(cls, values):  # type: ignore
                return func(cls, values)

            return wrapper

    return decorator


@dataclass(frozen=True)
class QueryConnectorConfig:
    # If using OVERWRITE, big query connector will first write to a staging
    # table followed by overwriting to the destination table.
    # When using APPEND, it is strongly recommended to use
    # MetadataField.JOB_TIMESTAMP to indicate the timestamp of the job.
    write_mode: WriteMode = WriteMode.APPEND


@dataclass(frozen=True)
class BigQueryOutputConfig(QueryConnectorConfig):
    pass


@dataclass(frozen=True)
class SnowflakeConnectorConfig(QueryConnectorConfig):
    pass


CONNECTOR_CONFIG_MAPPING = {
    'BigQueryConnector': BigQueryOutputConfig,
    'SnowflakeConnector': SnowflakeConnectorConfig,
    # 'DatabricksConnector': DatabricksOutputConfig,
    # 'S3Connector': S3OutputConfig,
}


@pydantic_dataclass(frozen=True, config={'arbitrary_types_allowed': True})
class OutputConfig:
    """Output configuration associated with a Batch Prediction Job.
    Specifies the output types and optionally output data source
    configuration.

    Args:
        output_types(`Set[str]`): The types of outputs that should be produced
            by the prediction job. Can include either ``'predictions'``,
            ``'embeddings'``, or both.
        output_connector(`Connector` or None): The output data source that Kumo
                should write batch predictions to, if it is None,
                produce local download output only.
        output_table_name(`str` or `Tuple[str, str]` or None): The name of the
            table in the output data source
            that Kumo should write batch predictions to. In the case of
            a Databricks connector, this should be a tuple of two strings:
            the schema name and the output prediction table name.
        output_metadata_fields(`List[MetadataField]` or None): Any additional
            metadata fields to include as new columns in the produced
            ``'predictions'`` output. Currently, allowed options are
            ``JOB_TIMESTAMP`` and ``ANCHOR_TIMESTAMP``.
        connector_specific_config(`QueryConnectorConfig` or None): The custom
            connector specific output config for predictions, for
            example whether to append or overwrite existing table.
    """
    output_types: Set[str]
    output_connector: Optional[Connector] = None
    output_table_name: Optional[Union[str, Tuple]] = None
    output_metadata_fields: Optional[List[MetadataField]] = None
    connector_specific_config: Optional[Union[
        BigQueryOutputConfig,
        SnowflakeConnectorConfig,
    ]] = None

    @compatible_field_validator('connector_specific_config')
    def validate_connector_config(cls, v: Any, values: Dict) -> Any:
        """Validate the connector specific output config. Raises ValueError if
        there is a mismatch between the connector type and the config type.
        """
        # Skip validation if no connector or no specific config
        if values.get('output_connector') is None or v is None:
            return v

        connector_type = type(values['output_connector']).__name__
        expected_config_type = CONNECTOR_CONFIG_MAPPING.get(connector_type)

        # If we don't have a mapping for this connector type, it doesn't
        # support specific configs yet
        if expected_config_type is None:
            raise ValueError(
                f"Connector type '{connector_type}' does not support "
                f"specific output configurations")

        # Check if the provided config is of the correct type
        if not isinstance(v, expected_config_type):
            raise ValueError(
                f"Connector type '{connector_type}' requires output "
                f"config of type '{expected_config_type.__name__}', but "
                f"got '{type(v).__name__}'")

        return v


@pydantic_dataclass(frozen=True, config={'arbitrary_types_allowed': True})
class TrainingTableExportConfig(OutputConfig):
    """Export configuration associated with a Training Table.

    Args:
        output_types(`Set[str]`): The artifact to export from the training
            table job. Currently only `'training_table'` is supported.
            Which exports the full training table to the output connector.
        output_connector(`Connector`): The output data source that Kumo should
            write training table artifacts to.
        output_table_name(str): The name of the table in the output data source
            that Kumo should write batch predictions to. In the case of
            a Databricks connector, this should be a tuple of two strings:
            the schema name and the output prediction table name.
        connector_specific_config(QueryConnectorConfig or None):
            Defines custom connector specific output
            for example whether to append or overwrite
            existing table. This is currently only supported for BigQuery and
            Snowflake.
    """
    output_connector: Connector
    output_table_name: str

    @compatible_model_validator()
    def validate(cls, values: Dict) -> Dict:
        output_types = values.get('output_types')
        if output_types != {'training_table'}:
            raise ValueError("output_type must be set(['training_table'])"
                             f" (got {output_types})")
        if values.get('output_connector') is None:
            raise ValueError("output_connector is required")
        if values.get('output_table_name') is None:
            raise ValueError("output_table_name is required")
        if values.get('output_metadata_fields') is not None:
            raise ValueError(
                "output_metadata_fields is not supported for training"
                "table export")
        return values



import asyncio
import concurrent
import logging
from io import StringIO
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    overload,
)

import pandas as pd
from kumoapi.jobs import JobStatus
from kumoapi.source_table import (
    DataSourceType,
    LLMRequest,
    SourceColumn,
    SourceTableDataResponse,
    SourceTableType,
)
from kumoapi.table import TableDefinition
from typing_extensions import override

from kumoai import global_state
from kumoai.client.jobs import LLMJobId
from kumoai.exceptions import HTTPException
from kumoai.futures import KumoFuture, create_future

if TYPE_CHECKING:
    from kumoai.connector import Connector

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL_S = 20.0


class SourceTable:
    r"""A source table is a reference to a table stored behind a backing
    :class:`~kumoai.connector.Connector`. It can be used to examine basic
    information about raw data connected to Kumo, including a sample of the
    table's rows, basic statistics, and column data type information.

    Once you are ready to use a table as part of a
    :class:`~kumoai.graph.Graph`, you may create a :class:`~kumoai.graph.Table`
    object from this source table, which includes additional specifying
    information (including column semantic types and column constraint
    information).

    Args:
        name: The name of this table in the backing connector
        connector: The connector containing this table.

    .. note::
        Source tables can also be augmented with large language models to
        introduce contextual embeddings for language features. To do so, please
        consult :meth:`~kumoai.connector.SourceTable.add_llm`.

    Example:
        >>> import kumoai
        >>> connector = kumoai.S3Connector(root_dir='s3://...')  # doctest: +SKIP # noqa: E501
        >>> articles_src = connector['articles']  # doctest: +SKIP
        >>> articles_src = kumoai.SourceTable('articles', connector)  # doctest: +SKIP # noqa: E501
    """
    def __init__(self, name: str, connector: 'Connector') -> None:
        # TODO(manan): existence check, if not too expensive?
        self.name = name
        self.connector = connector

    # Metadata ################################################################

    @property
    def column_dict(self) -> Dict[str, SourceColumn]:
        r"""Returns the names of the columns in this table along with their
        :class:`SourceColumn` information.
        """
        return {col.name: col for col in self.columns}

    @property
    def columns(self) -> List[SourceColumn]:
        r"""Returns a list of the :class:`SourceColumn` metadata of the columns
        in this table.
        """
        resp: SourceTableDataResponse = self.connector._get_table_data(
            table_names=[self.name], sample_rows=0)[0]
        return resp.cols

    # Data Access #############################################################

    def head(self, num_rows: int = 5) -> pd.DataFrame:
        r"""Returns the first :obj:`num_rows` rows of this source table by
        reading data from the backing connector.

        Args:
            num_rows: The number of rows to select. If :obj:`num_rows` is
                larger than the number of available rows, all rows will be
                returned.

        Returns:
            The first :obj:`num_rows` rows of the source table as a
            :class:`~pandas.DataFrame`.
        """
        num_rows = int(num_rows)
        if num_rows <= 0:
            raise ValueError(
                f"'num_rows' must be an integer greater than 0; got {num_rows}"
            )
        try:
            resp = self.connector._get_table_data([self.name], num_rows)[0]

            # TODO(manan, siyang): consider returning `bytes` instead of `json`
            assert resp.sample_rows is not None
            return pd.read_json(StringIO(resp.sample_rows), orient='table')
        except TypeError as e:
            raise RuntimeError(f"Cannot read head of table {self.name}. "
                               "Please restart the kernel and try.") from e

    # Language Models #########################################################

    @overload
    def add_llm(
        self,
        model: str,
        api_key: str,
        template: str,
        output_dir: str,
        output_column_name: str,
        output_table_name: str,
        dimensions: Optional[int],
    ) -> 'SourceTable':
        pass

    @overload
    def add_llm(
        self,
        model: str,
        api_key: str,
        template: str,
        output_dir: str,
        output_column_name: str,
        output_table_name: str,
        dimensions: Optional[int],
        *,
        non_blocking: Literal[False],
    ) -> 'SourceTable':
        pass

    @overload
    def add_llm(
        self,
        model: str,
        api_key: str,
        template: str,
        output_dir: str,
        output_column_name: str,
        output_table_name: str,
        dimensions: Optional[int],
        *,
        non_blocking: Literal[True],
    ) -> 'LLMSourceTableFuture':
        pass

    @overload
    def add_llm(
        self,
        model: str,
        api_key: str,
        template: str,
        output_dir: str,
        output_column_name: str,
        output_table_name: str,
        dimensions: Optional[int] = None,
        *,
        non_blocking: bool,
    ) -> Union['SourceTable', 'LLMSourceTableFuture']:
        pass

    def add_llm(
        self,
        model: str,
        api_key: str,
        template: str,
        output_dir: str,
        output_column_name: str,
        output_table_name: str,
        dimensions: Optional[int] = None,
        *,
        non_blocking: bool = False,
    ) -> Union['SourceTable', 'LLMSourceTableFuture']:
        r"""Experimental method which returns a new source table that
        includes a column computed via an LLM such as OpenAI embedding models.
        Please refer to the example script for more details.

        .. note::

            Current LLM embedding only works for :obj:`SourceTable` in s3.

        .. note::

            Your :obj:`api_key` will be encrypted once we received it and
            it's only decrypted just before we call the OpenAI text embeddings.

        .. note::
            Please keep track of the token usage in the `OpenAI Dashboard
            <https://platform.openai.com/usage/activity>`_. If number of
            tokens in the data exceeds the limit, the backend will raise
            an error and no result will be produced.

        .. warning::

            This method only supports text embedding with data that has less
            than ~6 million tokens. Number of tokens is estimated by following
            `this guide <https://help.openai.com/en/articles/4936856-what-are-tokens-and-how-to-count-them>`_.

        .. warning::

            This method is still experimental. Please consult with your Kumo
            POC before using it.

        Args:
            model: The LLM model name, *e.g.*, OpenAI's
                :obj:`"text-embedding-3-small"`.
            api_key: The API key to call the LLM service.
            template: A template string to be put into the LLM. For example,
                :obj:`"{A1} and {A2}"` will fuse columns :obj:`A1` and
                :obj:`A2` into a single string.
            output_dir: The S3 directory to store the output.
            output_column_name: The output column name for the LLM.
            output_table_name: The output table name.
            dimensions: The desired LLM embedding dimension.
            non_blocking: Whether making this function non-blocking.

        Example:
            >>> import kumoai
            >>> connector = kumoai.S3Connector(root_dir='s3://...')  # doctest: +SKIP # noqa: E501
            >>> articles_src = connector['articles']  # doctest: +SKIP
            >>> articles_src_future = \
                connector["articles"].add_llm(
                    model="text-embedding-3-small",
                    api_key=YOUR_OPENAI_API_KEY,
                    template=("The product {prod_name} in the {section_name} section"
                              "is categorized as {product_type_name} "
                              "and has following description: {detail_desc}"),
                    output_dir=YOUR_OUTPUT_DIR,
                    output_column_name="embedding_column",
                    output_table_name="articles_emb",
                    dimensions=256,
                    non_blocking=True,
                )
            >>> articles_src_future.status()  # doctest: +SKIP
            >>> articles_src_future.cancel()  # doctest: +SKIP
            >>> articles_src = articles_src_future.result()  # doctest: +SKIP
        """  # noqa
        if global_state.is_spcs:
            raise NotImplementedError("add_llm is not available on Snowflake")
        source_table_type = self._to_api_source_table()
        req = LLMRequest(
            source_table_type=source_table_type,
            template=template,
            model=model,
            model_api_key=api_key,
            output_dir=output_dir,
            output_column_name=output_column_name,
            output_table_name=output_table_name,
            dimensions=dimensions,
        )
        api = global_state.client.llm_job_api
        resp: LLMJobId = api.create(req)
        logger.info(f"LLMJobId: {resp}")
        source_table_future = LLMSourceTableFuture(resp, output_table_name,
                                                   output_dir)
        if non_blocking:
            return source_table_future
        # TODO (zecheng): Add attach for text embedding
        return source_table_future.result()

    # Persistence #############################################################

    def _to_api_source_table(self) -> SourceTableType:
        r"""Cast this source table as an object of type :obj:`SourceTableType`
        for use with the public REST API.
        """
        # TODO(manan): this is stupid, and is necessary because the s3_validate
        # method in Kumo core does not properly return directories. So, we have
        # to explicitly handle this ourselves here...
        try:
            return self.connector._get_table_config(self.name).source_table
        except HTTPException:
            name = self.name.rsplit('.', maxsplit=1)[0]
            out = self.connector._get_table_config(name).source_table
            self.name = name
            return out

    @staticmethod
    def _from_api_table_definition(
            table_definition: TableDefinition) -> 'SourceTable':
        r"""Constructs a :class:`SourceTable` from a
        :class:`kumoapi.table.TableDefinition`.
        """
        from kumoai.connector import (
            BigQueryConnector,
            DatabricksConnector,
            FileUploadConnector,
            S3Connector,
            SnowflakeConnector,
        )
        from kumoai.connector.s3_connector import S3URI
        source_type = table_definition.source_table.data_source_type
        connector: Connector
        if source_type == DataSourceType.S3:
            connector_id = table_definition.source_table.connector_id
            if connector_id in {
                    'parquet_upload_connector', 'csv_upload_connector'
            }:
                # File upload:
                connector = FileUploadConnector(
                    file_type=('parquet' if connector_id ==
                               'parquet_upload_connector' else 'csv'))
                table_name = table_definition.source_table.source_table_name
            else:
                if connector_id is not None:
                    connector = S3Connector(root_dir=None,
                                            _connector_id=connector_id)
                    table_name = (
                        table_definition.source_table.source_table_name)
                else:
                    table_path = S3URI(table_definition.source_table.s3_path)
                    connector = S3Connector(root_dir=table_path.root_dir)
                    # Strip suffix, since Kumo always takes care of that:
                    table_name = table_path.object_name.rsplit(
                        '.', maxsplit=1)[0]
        elif source_type == DataSourceType.SNOWFLAKE:
            connector_api = global_state.client.connector_api
            connector_resp = connector_api.get(
                table_definition.source_table.snowflake_connector_id)
            assert connector_resp is not None
            connector_cfg = connector_resp.config
            connector = SnowflakeConnector(
                name=connector_cfg.name,
                account=connector_cfg.account,
                warehouse=connector_cfg.warehouse,
                database=connector_cfg.database,
                schema_name=connector_cfg.schema_name,
                credentials=None,  # should be in env; do not load from DB.
                _bypass_creation=True,
            )
            table_name = table_definition.source_table.table
        elif source_type == DataSourceType.DATABRICKS:
            connector_api = global_state.client.connector_api
            connector_resp = connector_api.get(
                table_definition.source_table.databricks_connector_id)
            assert connector_resp is not None
            connector_cfg = connector_resp.config
            connector = DatabricksConnector(
                name=connector_cfg.name,
                host=connector_cfg.host,
                cluster_id=connector_cfg.cluster_id,
                warehouse_id=connector_cfg.warehouse_id,
                catalog=connector_cfg.catalog,
                credentials=None,  # should be in env; do not load from DB.
                _bypass_creation=True,
            )
            table_name = table_definition.source_table.table
        elif source_type == DataSourceType.BIGQUERY:
            connector_api = global_state.client.connector_api
            connector_resp = connector_api.get(
                table_definition.source_table.bigquery_connector_id)
            assert connector_resp is not None
            connector_cfg = connector_resp.config
            connector = BigQueryConnector(
                name=connector_cfg.name,
                project_id=connector_cfg.project_id,
                dataset_id=connector_cfg.dataset_id,
                credentials=None,  # should be in env; do not load from DB.
                _bypass_creation=True,
            )
            table_name = table_definition.source_table.table_name
        else:
            raise NotImplementedError()

        return SourceTable(table_name, connector)

    # Class properties ########################################################

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name})'


class SourceTableFuture(KumoFuture[SourceTable]):
    r"""A representation of an on-going :class:`SourceTable` generation
    process.
    """
    def __init__(
        self,
        job_id: LLMJobId,
        table_name: str,
        output_dir: str,
    ) -> None:
        self.job_id = job_id
        self._fut: concurrent.futures.Future = create_future(
            _poll(job_id, table_name, output_dir))

    @override
    def result(self) -> SourceTable:
        return self._fut.result()

    @override
    def future(self) -> 'concurrent.futures.Future[SourceTable]':
        return self._fut

    def status(self) -> JobStatus:
        return _get_status(self.job_id)


class LLMSourceTableFuture(SourceTableFuture):
    r"""A representation of an on-going :class:`SourceTable`
    generation process for LLM. This class inherits from the
    :class:`SourceTableFuture` with some functions specific
    to LLM job.
    """
    def __init__(
        self,
        job_id: LLMJobId,
        table_name: str,
        output_dir: str,
    ) -> None:
        super().__init__(job_id, table_name, output_dir)

    def cancel(self) -> JobStatus:
        r"""Cancel the LLM job."""
        api = global_state.client.llm_job_api
        return api.cancel(self.job_id)


def _get_status(job_id: str) -> JobStatus:
    api = global_state.client.llm_job_api
    resource: JobStatus = api.get(job_id)
    return resource


async def _poll(job_id: str, table_name: str, output_dir: str) -> SourceTable:
    status = _get_status(job_id)
    while not status.is_terminal:
        await asyncio.sleep(_DEFAULT_INTERVAL_S)
        status = _get_status(job_id)

    if status != JobStatus.DONE:
        raise RuntimeError(f"LLM job {job_id} failed with "
                           f"job status {status}.")

    from kumoai.connector import S3Connector
    connector = S3Connector(root_dir=output_dir)

    return connector[table_name]



from http import HTTPStatus
from typing import Any, Dict, Optional

from kumoapi.common import ValidationResponse
from kumoapi.data_snapshot import GraphSnapshotID, GraphSnapshotResource
from kumoapi.graph import (
    EdgeHealthResponse,
    GraphDefinition,
    GraphResource,
    GraphValidationRequest,
)
from kumoapi.json_serde import to_json_dict

from kumoai.client import KumoClient
from kumoai.client.endpoints import GraphEndpoints
from kumoai.client.utils import (
    parse_id_response,
    parse_response,
    raise_on_error,
)

GraphID = str


class GraphAPI:
    r"""Typed API definition for Kumo graph definition."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client

    def create_graph(
        self,
        graph_def: GraphDefinition,
        *,
        name_alias: Optional[str] = None,
        force_rename: bool = False,
    ) -> GraphID:
        r"""Creates a Graph (metadata definition) resource object in Kumo."""
        params: Dict[str, Any] = {'force_rename': force_rename}
        if name_alias:
            params['name_alias'] = name_alias
        resp = self._client._request(
            GraphEndpoints.create,
            params=params,
            json=to_json_dict(graph_def),
        )
        raise_on_error(resp)
        return parse_id_response(resp)

    def get_graph_if_exists(
        self,
        graph_id_or_name: str,
    ) -> Optional[GraphResource]:
        resp = self._client._request(
            GraphEndpoints.get.with_id(graph_id_or_name))
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise_on_error(resp)
        return parse_response(GraphResource, resp)

    def create_snapshot(
        self,
        graph_id: GraphID,
        *,
        refresh_source: bool = False,
    ) -> GraphSnapshotID:
        params: Dict[str, Any] = {
            'graph_id': graph_id,
            'refresh_source': refresh_source
        }
        resp = self._client._request(GraphEndpoints.create_snapshot,
                                     params=params)
        raise_on_error(resp)
        return GraphSnapshotID(parse_id_response(resp))

    def get_snapshot(
        self,
        snapshot_id: GraphSnapshotID,
    ) -> GraphSnapshotResource:
        resp = self._client._request(
            GraphEndpoints.get_snapshot.with_id(snapshot_id))
        raise_on_error(resp)
        return parse_response(GraphSnapshotResource, resp)

    def get_snapshot_if_exists(
        self,
        snapshot_id: GraphSnapshotID,
    ) -> Optional[GraphSnapshotResource]:
        resp = self._client._request(
            GraphEndpoints.get_snapshot.with_id(snapshot_id))
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise_on_error(resp)
        return parse_response(GraphSnapshotResource, resp)

    def get_edge_stats(
        self,
        graph_snapshot_id: GraphSnapshotID,
    ) -> EdgeHealthResponse:
        r"""Fetches edge statistics given a snapshot id"""
        resp = self._client._request(
            GraphEndpoints.get_edge_stats.with_id(graph_snapshot_id))
        raise_on_error(resp)
        return parse_response(EdgeHealthResponse, resp)

    def validate_graph(
        self,
        request: GraphValidationRequest,
    ) -> ValidationResponse:
        response = self._client._request(GraphEndpoints.validate,
                                         json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(ValidationResponse, response)

    def infer_links(self, graph: GraphDefinition) -> GraphDefinition:
        resp = self._client._request(GraphEndpoints.infer_links,
                                     json=to_json_dict(graph))
        raise_on_error(resp)
        return parse_response(GraphDefinition, resp)



from .id import contains_id
from .timestamp import contains_timestamp
from .categorical import contains_categorical
from .multicategorical import contains_multicategorical

__all__ = [
    'contains_id',
    'contains_timestamp',
    'contains_categorical',
    'contains_multicategorical',
]
